from __future__ import annotations

import json
import time
from typing import TYPE_CHECKING, Any

import numpy as np
import py_scibec
import py_scibec_openapi_client.models as py_scibec_models

from bec_lib import MessageEndpoints, bec_logger
from bec_lib.serialization import json_ext

logger = bec_logger.logger

if TYPE_CHECKING:
    from bec_lib import messages
    from bec_server.scihub.scibec import SciBecConnector


class SciBecMetadataHandler:
    """
    The SciBecMetadataHandler class is responsible for handling metadata sent to SciBec.
    """

    MAX_DATA_SIZE = 1e6  # max data size for the backend; currently set to 1 MB

    def __init__(self, scibec_connector: SciBecConnector) -> None:
        self.scibec_connector = scibec_connector
        self._scan_status_register = None
        self._start_scan_subscription()
        self._file_subscription = None
        self._start_file_subscription()

    def _start_scan_subscription(self):
        self._scan_status_register = self.scibec_connector.connector.register(
            MessageEndpoints.scan_status(), cb=self._handle_scan_status, parent=self
        )

    def _start_file_subscription(self):
        self._file_subscription = self.scibec_connector.connector.register(
            MessageEndpoints.file_content(), cb=self._handle_file_content, parent=self
        )

    @staticmethod
    def _handle_scan_status(msg, *, parent, **_kwargs) -> None:
        msg = msg.value
        try:
            scan = parent.update_scan_status(msg)
        except Exception as exc:
            logger.exception(f"Failed to update scan status: {exc}")
            logger.warning("Failed to write to SciBec")
            return

        # if msg.content["status"] != "open":
        #     parent.update_event_data(scan)

    def update_scan_status(self, msg: messages.ScanStatusMessage) -> dict:
        """
        Update the scan status in SciBec

        Args:
            msg(messages.ScanStatusMessage): The message containing the scan data

        Returns:
            dict: The updated scan data
        """
        scibec = self.scibec_connector.scibec
        if not scibec:
            return
        scibec_info = self.scibec_connector.scibec_info
        experiment_id = scibec_info["activeExperiment"].id
        # session_id = scibec_info["activeSession"][0]["id"]
        # experiment_id = scibec_info["activeSession"][0]["experimentId"]
        logger.debug(f"Received new scan status {msg}")
        scan_filter = py_scibec.bec.ScanFilterWhere(where={"scanId": msg.content["scan_id"]})
        scan = scibec.scan.scan_controller_find(scan_filter)
        if not scan:
            info = msg.content["info"]
            dataset_number = info.get("dataset_number")
            dataset_filter = py_scibec.bec.DatasetFilterWhere(
                where={"number": dataset_number, "experimentId": experiment_id}
            )
            dataset = scibec.dataset.dataset_controller_find(dataset_filter)
            if dataset:
                dataset = dataset[0]
            else:
                new_dataset = py_scibec.bec.NewDataset(
                    **{
                        "readACL": scibec_info["activeExperiment"].read_acl,
                        "writeACL": scibec_info["activeExperiment"].read_acl,
                        "owner": scibec_info["activeExperiment"].owner,
                        "number": dataset_number,
                        "experimentId": experiment_id,
                        "name": info.get("dataset_name", ""),
                    }
                )
                dataset = scibec.dataset.dataset_controller_create(new_dataset)

            new_scan = py_scibec.bec.NewScan(
                **{
                    "readACL": scibec_info["activeExperiment"].read_acl,
                    "writeACL": scibec_info["activeExperiment"].read_acl,
                    "owner": scibec_info["activeExperiment"].owner,
                    "scanType": info.get("scan_name", ""),
                    "scanId": info.get("scan_id", ""),
                    "queueId": info.get("queue_id", ""),
                    "requestId": info.get("RID", ""),
                    "exitStatus": msg.content["status"],
                    # "queue": info.get("stream", ""),
                    "metadata": info,
                    # "sessionId": session_id,
                    "datasetId": dataset.id,
                    "scanNumber": info.get("scan_number", 0),
                }
            )
            scan = scibec.scan.scan_controller_create(new_scan)
            # scan = scibec.add_scan(scan_data)
        else:
            info = msg.content["info"]
            scan = scibec.scan.scan_controller_update_by_id(
                id=scan[0].id,
                scan_partial=py_scibec.bec.ScanPartial(
                    metadata=info, exitStatus=msg.content["status"]
                ),
            )
        return scan

    @staticmethod
    def _handle_file_content(msg, *, parent, **_kwargs) -> None:
        msg = msg.value
        try:
            logger.debug(f"Received new file content {msg}")
            if not msg.content["data"]:
                return
            parent.update_scan_data(**msg.content)
        except Exception as exc:
            logger.exception(f"Failed to update scan data: {exc}")
            logger.warning("Failed to write to SciBec")
            return

    def serialize_special_data(self, data: Any) -> dict:
        """
        Serialize special data in the scan data.
        This method is recursively called for each key in the data dictionary and
        checks if the data is serializable. If not, the data is serialized using
        the json_ext.dumps method.

        Args:
            data(dict): The scan data

        Returns:
            dict: The serialized scan data
        """
        if isinstance(data, (int, float, str, bool)):
            return data
        if isinstance(data, dict):
            return {key: self.serialize_special_data(value) for key, value in data.items()}
        if isinstance(data, list):
            return [self.serialize_special_data(item) for item in data]
        if isinstance(data, tuple):
            return tuple(self.serialize_special_data(item) for item in data)
        if isinstance(data, set):
            return {self.serialize_special_data(item) for item in data}
        if isinstance(data, np.generic):
            return self.serialize_special_data(data.tolist())
        return json_ext.dumps(data)

    def update_scan_data(self, file_path: str, data: dict):
        """
        Update the scan data in SciBec

        Args:
            file_path(str): The path to the original NeXuS file
            data(dict): The scan data
        """
        scibec = self.scibec_connector.scibec
        if not scibec:
            return
        scan_filter = py_scibec.bec.ScanFilterWhere(where={"scanId": data["metadata"]["scan_id"]})
        scan = scibec.scan.scan_controller_find(scan_filter)
        if not scan:
            logger.warning(
                f"Could not find scan with scan_id {data['metadata']['scan_id']}. Cannot write scan"
                " data to SciBec."
            )
            return
        scan = scan[0]
        data_bec = self.serialize_special_data(data)
        data_size = len(json.dumps(data_bec)) / self.MAX_DATA_SIZE

        start = time.time()
        logger.info(f"Data size: {data_size} MB")
        if data_size > 1:
            logger.info(
                f"Data size is larger than {self.MAX_DATA_SIZE/1e6} MB. Splitting data into chunks."
            )
            self._write_scan_data_chunks(file_path, data_bec, scan)
        else:
            new_scan_data = py_scibec.bec.NewScanData(
                **{
                    "readACL": scan.read_acl,
                    "writeACL": scan.read_acl,
                    "owner": scan.owner,
                    "scanId": scan.id,
                    "filePath": file_path,
                    "data": data_bec,
                }
            )
            scibec.scan_data.scan_data_controller_create_many(new_scan_data)
        logger.info(
            f"Wrote scan data to SciBec for scan_id {data['metadata']['scan_id']} in {time.time() - start} seconds."
        )

    def _write_scan_data_chunks(
        self, file_path: str, data_bec: dict, scan: py_scibec_models.ScanWithRelations
    ):
        """
        Write the scan data to SciBec in chunks. This method is called if the scan data is larger
        than 1 MB. The method loops through all keys in the data dictionary and creates chunks of
        max 1 MB size. The chunks are then written to SciBec.

        Args:
            file_path(str): The path to the original NeXuS file
            data_bec(dict): The serialized scan data
            scan(dict): The scan data
        """
        scibec = self.scibec_connector.scibec
        if not scibec:
            return
        chunk = {}
        for key, value in data_bec.items():
            if len(json.dumps({key: value})) > self.MAX_DATA_SIZE:
                logger.warning(
                    f"Data size of key {key} is larger than {self.MAX_DATA_SIZE/1e6} MB. Cannot write this key to SciBec."
                )
                continue
            if len(json.dumps(chunk)) + len(json.dumps({key: value})) > self.MAX_DATA_SIZE:
                new_scan_data = py_scibec.bec.NewScanData(
                    **{
                        "readACL": scan.read_acl,
                        "writeACL": scan.read_acl,
                        "owner": scan.owner,
                        "scanId": scan.id,
                        "filePath": file_path,
                        "data": chunk,
                    }
                )
                scibec.scan_data.scan_data_controller_create_many(new_scan_data)
                chunk = {}
            chunk[key] = value

        # Write the last chunk
        if chunk:
            new_scan_data = py_scibec.bec.NewScanData(
                **{
                    "readACL": scan.read_acl,
                    "writeACL": scan.read_acl,
                    "owner": scan.owner,
                    "scanId": scan.id,
                    "filePath": file_path,
                    "data": chunk,
                }
            )
            scibec.scan_data.scan_data_controller_create_many(new_scan_data)

    def shutdown(self):
        """
        Shutdown the metadata handler
        """
        if self._scan_status_register:
            self._scan_status_register.shutdown()
        if self._file_subscription:
            self._file_subscription.shutdown()



"""
Scan Report class that provides a convenient way to access the status of a scan request. It is typically
the return value of a scan request. 
"""

from __future__ import annotations

import time
from math import inf
from typing import TYPE_CHECKING

from bec_lib import messages
from bec_lib.bec_errors import ScanAbortion
from bec_lib.endpoints import MessageEndpoints

if TYPE_CHECKING:
    from bec_lib.client import BECClient
    from bec_lib.queue_items import QueueItem


class ScanReport:
    """Scan Report class that provides a convenient way to access the status of a scan request."""

    def __init__(self) -> None:
        self._client = None
        self.request = None
        self._queue_item = None

    @classmethod
    def from_request(
        cls, request: messages.ScanQueueMessage, client: BECClient = None
    ) -> ScanReport:
        """
        Create a ScanReport from a request

        Args:
            request (messages.ScanQueueMessage): request to create the report from
            client (BECClient, optional): BECClient instance. Defaults to None.

        Returns:
            ScanReport: ScanReport instance
        """
        scan_report = cls()
        scan_report._client = client

        client.queue.request_storage.update_with_request(request)
        scan_report.request = client.queue.request_storage.find_request_by_ID(
            request.metadata["RID"]
        )
        return scan_report

    @property
    def scan(self):
        """get the scan item"""
        return self.request.scan

    @property
    def status(self):
        """returns the current status of the request"""
        scan_type = self.request.request.content["scan_type"]
        status = self.queue_item.status
        if scan_type == "mv" and status == "COMPLETED":
            return "COMPLETED" if self._get_mv_status() else "RUNNING"
        return self.queue_item.status

    @property
    def queue_item(self):
        """get the queue item"""
        if not self._queue_item:
            self._queue_item = self._get_queue_item(timeout=10)
        return self._queue_item

    def _get_queue_item(self, timeout=None) -> QueueItem:
        """
        get the queue item from the queue storage

        Args:
            timeout (float, optional): timeout in seconds. Defaults to None.
        """
        timeout = timeout if timeout is not None else inf
        queue_item = None
        elapsed_time = 0
        sleep_time = 0.1
        while not queue_item:
            queue_item = self._client.queue.queue_storage.find_queue_item_by_requestID(
                self.request.requestID
            )
            elapsed_time += sleep_time
            time.sleep(sleep_time)
            if elapsed_time > timeout:
                raise TimeoutError
        return queue_item

    def _get_mv_status(self) -> bool:
        """get the status of a move request"""
        motors = list(self.request.request.content["parameter"]["args"].keys())
        request_status = self._client.device_manager.connector.lrange(
            MessageEndpoints.device_req_status_container(self.request.requestID), 0, -1
        )
        if len(request_status) == len(motors):
            return True
        return False

    def wait(self, timeout: float = None) -> ScanReport:
        """
        wait for the request to complete

        Args:
            timeout (float, optional): timeout in seconds. Defaults to None.

        Raises:
            TimeoutError: if the timeout is reached

        Returns:
            ScanReport: ScanReport instance
        """
        sleep_time = 0.1
        scan_type = self.request.request.content["scan_type"]
        try:
            if scan_type == "mv":
                self._wait_move(timeout, sleep_time)
            else:
                self._wait_scan(timeout, sleep_time)
        except KeyboardInterrupt as exc:
            self._client.queue.request_scan_abortion()
            raise ScanAbortion("Aborted by user.") from exc

        return self

    def _check_timeout(self, timeout: float = None, elapsed_time: float = 0) -> None:
        """
        check if the timeout is reached

        Args:
            timeout (float, optional): timeout in seconds. Defaults to None.
            elapsed_time (float, optional): elapsed time in seconds. Defaults to 0.

        """
        if timeout is None:
            return
        if elapsed_time > timeout:
            raise TimeoutError(
                f"Timeout reached while waiting for request to complete. Timeout: {timeout} s."
            )

    def _wait_move(self, timeout: float = None, sleep_time: float = 0.1) -> None:
        """
        wait for a move request to complete

        Args:
            timeout (float, optional): timeout in seconds. Defaults to None.
            sleep_time (float, optional): sleep time in seconds. Defaults to 0.1.

        """
        elapsed_time = 0
        while True:
            if self._get_mv_status():
                break
            self._client.alarm_handler.raise_alarms()
            time.sleep(sleep_time)
            elapsed_time += sleep_time
            self._check_timeout(timeout, elapsed_time)

    def _wait_scan(self, timeout: float = None, sleep_time: float = 0.1) -> None:
        """
        wait for a scan request to complete

        Args:
            timeout (float, optional): timeout in seconds. Defaults to None.
            sleep_time (float, optional): sleep time in seconds. Defaults to 0.1.
        """
        elapsed_time = 0
        while True:
            if self.status == "COMPLETED":
                break
            if self.status == "STOPPED":
                raise ScanAbortion
            self._client.callbacks.poll()
            time.sleep(sleep_time)
            elapsed_time += sleep_time
            self._check_timeout(timeout, elapsed_time)

    def __str__(self) -> str:
        separator = "--" * 10
        details = f"\tStatus: {self.status}\n"
        if self.scan:
            details += self.scan.describe()
        return f"ScanReport:\n{separator}\n{details}"



import sys
from string import Template

from bec_server.bec_server_utils.subprocess_launch import subprocess_start, subprocess_stop
from bec_server.bec_server_utils.tmux_launch import tmux_start, tmux_stop


class bcolors:
    """
    Colors for the terminal output.
    """

    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


class ServiceHandler:
    """
    Service handler for the BEC server. This class is used to start, stop and restart the BEC server.
    Depending on the platform, the server is launched in a tmux session or in an iTerm2 session.
    """

    SERVICES = {
        "scan_server": {"path": Template("$base_path/scan_server"), "command": "bec-scan-server"},
        "scan_bundler": {
            "path": Template("$base_path/scan_bundler"),
            "command": "bec-scan-bundler",
        },
        "device_server": {
            "path": Template("$base_path/device_server"),
            "command": "bec-device-server",
        },
        "file_writer": {"path": Template("$base_path/file_writer"), "command": "bec-file-writer"},
        "scihub": {"path": Template("$base_path/scihub"), "command": "bec-scihub"},
        "data_processing": {"path": Template("$base_path/data_processing"), "command": "bec-dap"},
    }

    def __init__(self, bec_path: str, config_path: str, no_tmux: bool = False):
        """

        Args:
            bec_path (str): Path to the BEC source code
            config_path (str): Path to the config file

        """
        self.bec_path = bec_path
        self.config_path = config_path
        self.no_tmux = no_tmux

        self._detect_available_interfaces()

    def _detect_available_interfaces(self):
        if self.no_tmux:
            self.interface = None
            return
        # check if we are on MacOS and if so, check if we have iTerm2 installed
        if sys.platform == "darwin":
            try:
                import iterm2
            except ImportError:
                self.interface = "tmux"
            else:
                self.interface = "iterm2"

        # if we are not on MacOS, we can only use tmux
        else:
            self.interface = "tmux"

    def start(self):
        """
        Start the BEC server using the available interface.
        """
        if self.interface == "tmux":
            print("Starting BEC server using tmux...")
            tmux_start(self.bec_path, self.config_path, self.SERVICES)
            print(
                f"{bcolors.OKCYAN}{bcolors.BOLD}Use `tmux attach -t bec` to attach to the BEC server. Once connected, use `ctrl+b d` to detach again."
            )
        elif self.interface == "iterm2":
            pass
        else:
            # no tmux
            return subprocess_start(self.bec_path, self.config_path, self.SERVICES)

    def stop(self):
        """
        Stop the BEC server using the available interface.
        """
        print("Stopping BEC server...")
        if self.interface == "tmux":
            tmux_stop()
        elif self.interface == "iterm2":
            pass
        else:
            subprocess_stop()

    def restart(self):
        """
        Restart the BEC server using the available interface.
        """
        print("Restarting BEC server...")
        self.stop()
        self.start()



import collections
import threading
import time
import traceback
from collections.abc import Callable
from concurrent.futures import ThreadPoolExecutor

from bec_lib import BECService, BECStatus, DeviceManagerBase, MessageEndpoints, bec_logger, messages
from bec_lib.connector import ConnectorBase

from .bec_emitter import BECEmitter
from .bluesky_emitter import BlueskyEmitter

logger = bec_logger.logger


class ScanBundler(BECService):
    def __init__(self, config, connector_cls: ConnectorBase) -> None:
        super().__init__(config, connector_cls, unique_service=True)

        self.device_manager = None
        self._start_device_manager()
        self.connector.register(
            patterns=MessageEndpoints.device_read("*"),
            cb=self._device_read_callback,
            name="device_read_register",
        )
        self.connector.register(
            MessageEndpoints.scan_queue_status(),
            cb=self._scan_queue_callback,
            group_id="scan_bundler",
            name="scan_queue_register",
        )
        self.connector.register(
            MessageEndpoints.scan_status(),
            cb=self._scan_status_callback,
            group_id="scan_bundler",
            name="scan_status_register",
        )

        self.sync_storage = {}
        self.monitored_devices = {}
        self.baseline_devices = {}
        self.device_storage = {}
        self.scan_motors = {}
        self.readout_priority = {}
        self.storage_initialized = set()
        self.current_queue = None
        self.executor = ThreadPoolExecutor(max_workers=4)
        self.executor_tasks = collections.deque(maxlen=100)
        self.scan_id_history = collections.deque(maxlen=10)
        self._lock = threading.Lock()
        self._emitter = []
        self._initialize_emitters()
        self.status = BECStatus.RUNNING

    def _initialize_emitters(self):
        self._emitter = [BECEmitter(self), BlueskyEmitter(self)]

    def run_emitter(self, emitter_method: Callable, *args, **kwargs):
        for emi in self._emitter:
            try:
                getattr(emi, emitter_method)(*args, **kwargs)
            except Exception:
                content = traceback.format_exc()
                logger.error(f"Failed to run emitter: {content}")

    def _start_device_manager(self):
        self.wait_for_service("DeviceServer")
        self.device_manager = DeviceManagerBase(self)
        self.device_manager.initialize(self.bootstrap_server)

    def _device_read_callback(self, msg, **_kwargs):
        # pylint: disable=protected-access
        dev = msg.topic.split(MessageEndpoints.device_read("").endpoint)[-1]
        msgs = msg.value
        logger.debug(f"Received reading from device {dev}")
        if not isinstance(msgs, list):
            msgs = [msgs]
        task = self.executor.submit(self._add_device_to_storage, msgs, dev)
        self.executor_tasks.append(task)

    def _scan_queue_callback(self, msg, **_kwargs):
        msg = msg.value
        logger.trace(msg)
        self.current_queue = msg.content["queue"]["primary"].get("info")

    def _scan_status_callback(self, msg, **_kwargs):
        msg = msg.value
        self.handle_scan_status_message(msg)

    def handle_scan_status_message(self, msg: messages.ScanStatusMessage) -> None:
        """handle scan status messages"""
        logger.info(f"Received new scan status: {msg}")
        scan_id = msg.content["scan_id"]
        self.cleanup_storage()
        if scan_id not in self.sync_storage:
            self._initialize_scan_container(msg)
            if scan_id not in self.scan_id_history:
                self.scan_id_history.append(scan_id)
        if msg.content.get("status") != "open":
            self._scan_status_modification(msg)

    def _scan_status_modification(self, msg: messages.ScanStatusMessage):
        status = msg.content.get("status")
        if status not in ["closed", "aborted", "paused", "halted"]:
            logger.error(f"Unknown scan status {status}")
            return

        scan_id = msg.content.get("scan_id")
        if not scan_id:
            logger.warning(f"Received scan status update without scan_id: {msg}")
            return
        if self.sync_storage.get(scan_id):
            self.sync_storage[scan_id]["status"] = status
        else:
            self.sync_storage[scan_id] = {"info": {}, "status": status, "sent": set()}
            self.storage_initialized.add(scan_id)
            if scan_id not in self.scan_id_history:
                self.scan_id_history.append(scan_id)

    def _initialize_scan_container(self, scan_msg: messages.ScanStatusMessage):
        if scan_msg.content.get("status") != "open":
            return

        scan_id = scan_msg.content["scan_id"]
        scan_info = scan_msg.content["info"]
        scan_motors = list(set(self.device_manager.devices[m] for m in scan_info["scan_motors"]))
        self.scan_motors[scan_id] = scan_motors
        self.readout_priority[scan_id] = scan_info["readout_priority"]
        if scan_id not in self.storage_initialized:
            self.sync_storage[scan_id] = {"info": scan_info, "status": "open", "sent": set()}
            self.monitored_devices[scan_id] = {
                "devices": self.device_manager.devices.monitored_devices(
                    readout_priority=self.readout_priority[scan_id]
                ),
                "point_id": {},
            }
            self.baseline_devices[scan_id] = {
                "devices": self.device_manager.devices.baseline_devices(
                    readout_priority=self.readout_priority[scan_id]
                ),
                "done": {
                    dev.name: False
                    for dev in self.device_manager.devices.baseline_devices(
                        readout_priority=self.readout_priority[scan_id]
                    )
                },
            }
            self.storage_initialized.add(scan_id)
            self.run_emitter("on_init", scan_id)
            return

    def _step_scan_update(self, scan_id, device, signal, metadata):
        if "point_id" not in metadata:
            return
        with self._lock:
            dev = {device: signal}
            point_id = metadata["point_id"]
            monitored_devices = self.monitored_devices[scan_id]

            self.sync_storage[scan_id][point_id] = {
                **self.sync_storage[scan_id].get(point_id, {}),
                **dev,
            }

            if monitored_devices["point_id"].get(point_id) is None:
                monitored_devices["point_id"][point_id] = {
                    dev.name: False for dev in monitored_devices["devices"]
                }
            monitored_devices["point_id"][point_id][device] = True

            monitored_devices_completed = list(monitored_devices["point_id"][point_id].values())

            all_monitored_devices_completed = bool(
                all(monitored_devices_completed)
                and (
                    len(monitored_devices_completed)
                    == len(self.monitored_devices[scan_id]["devices"])
                )
            )

            if all_monitored_devices_completed and self.sync_storage[scan_id].get(point_id):
                self._update_monitor_signals(scan_id, point_id)
                self._send_scan_point(scan_id, point_id)

    def _fly_scan_update(self, scan_id, device, signal, metadata):
        if "point_id" not in metadata:
            logger.warning(
                f"Received device message from device {device} without point_id in metadata. {metadata}"
            )
            return
        with self._lock:
            dev = {device: signal}
            point_id = metadata["point_id"]
            logger.info(
                f"Received reading from device {device} for scan_id {scan_id} at point {point_id}."
            )
            if self.sync_storage[scan_id].get("info", {}).get("monitor_sync", "bec") == "bec":
                # For monitor sync with BEC, we use the point_id as the key for the sync_storage.
                monitored_devices = self.monitored_devices[scan_id]

                self.sync_storage[scan_id][point_id] = {
                    **self.sync_storage[scan_id].get(point_id, {}),
                    **dev,
                }

                if monitored_devices["point_id"].get(point_id) is None:
                    monitored_devices["point_id"][point_id] = {
                        dev.name: False for dev in monitored_devices["devices"]
                    }
                monitored_devices["point_id"][point_id][device] = True

                monitored_devices_completed = list(monitored_devices["point_id"][point_id].values())

                all_monitored_devices_completed = bool(
                    all(monitored_devices_completed)
                    and (
                        len(monitored_devices_completed)
                        == len(self.monitored_devices[scan_id]["devices"])
                    )
                )
                if all_monitored_devices_completed and self.sync_storage[scan_id].get(point_id):
                    self._update_monitor_signals(scan_id, point_id)
                    self._send_scan_point(scan_id, point_id)
            else:
                self.sync_storage[scan_id][point_id] = {
                    **self.sync_storage[scan_id].get(point_id, {}),
                    **signal,
                }

                if self.sync_storage[scan_id].get(point_id):
                    self._update_monitor_signals(scan_id, point_id)
                    self._send_scan_point(scan_id, point_id)

    def _baseline_update(self, scan_id, device, signal):
        with self._lock:
            dev = {device: signal}
            baseline_devices_status = self.baseline_devices[scan_id]["done"]
            baseline_devices_status[device] = True

            self.sync_storage[scan_id]["baseline"] = {
                **self.sync_storage[scan_id].get("baseline", {}),
                **dev,
            }

            if not all(status for status in baseline_devices_status.values()):
                return

            logger.info(f"Sending baseline readings for scan_id {scan_id}.")
            logger.debug("Baseline: ", self.sync_storage[scan_id]["baseline"])
            self.run_emitter("on_baseline_emit", scan_id)
            self.baseline_devices[scan_id]["done"] = {
                dev.name: False
                for dev in self.device_manager.devices.baseline_devices(
                    readout_priority=self.readout_priority[scan_id]
                )
            }

    def _wait_for_scan_id(self, scan_id, timeout_time=10):
        elapsed_time = 0
        while scan_id not in self.storage_initialized:
            msg = self.connector.get(MessageEndpoints.public_scan_info(scan_id))
            if msg and msg.content["scan_id"] == scan_id:
                self.handle_scan_status_message(msg)
            if scan_id in self.sync_storage:
                if self.sync_storage[scan_id]["status"] in ["closed", "aborted"]:
                    logger.info(
                        f"Received reading for {self.sync_storage[scan_id]['status']} scan {scan_id}."
                    )
                    return
            time.sleep(0.05)
            elapsed_time += 0.05
            if elapsed_time > timeout_time:
                raise TimeoutError(
                    f"Reached timeout whilst waiting for scan_id {scan_id} to appear."
                )

    def _add_device_to_storage(self, msgs, device, timeout_time=10):
        for msg in msgs:
            metadata = msg.metadata

            scan_id = metadata.get("scan_id")
            if not scan_id:
                logger.info("Received device message without scan_id")
                return

            signal = msg.content.get("signals")

            try:
                self._wait_for_scan_id(scan_id, timeout_time=timeout_time)
            except TimeoutError:
                logger.warning(f"Could not find a matching scan_id {scan_id} in sync_storage.")
                return

            if self.sync_storage[scan_id]["status"] in ["aborted", "closed"]:
                # check if the sync_storage has been initialized properly.
                # In case of post-scan initialization, scan info is not available
                if not self.sync_storage[scan_id]["info"].get("scan_type"):
                    return
            self.device_storage[device] = signal
            readout_priority = metadata.get("readout_priority")
            device_is_monitor_sync = self.sync_storage[scan_id]["info"]["monitor_sync"] == device
            dev_obj = self.device_manager.devices.get(device)
            if dev_obj in self.monitored_devices[scan_id]["devices"] or device_is_monitor_sync:
                if self.sync_storage[scan_id]["info"]["scan_type"] == "step":
                    self._step_scan_update(scan_id, device, signal, metadata)
                elif self.sync_storage[scan_id]["info"]["scan_type"] == "fly":
                    self._fly_scan_update(scan_id, device, signal, metadata)
                else:
                    raise RuntimeError(
                        f"Unknown scan type {self.sync_storage[scan_id]['info']['scan_type']}"
                    )
            elif readout_priority == "baseline":
                self._baseline_update(scan_id, device, signal)
            else:
                logger.warning(f"Received reading from unknown device {device}")

    def _update_monitor_signals(self, scan_id, point_id) -> None:
        if self.sync_storage[scan_id]["info"]["scan_type"] == "fly":
            # for fly scans, take all primary and monitor signals
            devices = self.monitored_devices[scan_id]["devices"]

            readings = self._get_last_device_readback(devices)

            for read, dev in zip(readings, devices):
                self.sync_storage[scan_id][point_id][dev.name] = read

    def _get_last_device_readback(self, devices: list) -> list:
        pipe = self.connector.pipeline()
        for dev in devices:
            self.connector.get(MessageEndpoints.device_readback(dev.name), pipe)
        return [msg.content["signals"] for msg in self.connector.execute_pipeline(pipe)]

    def cleanup_storage(self):
        """remove old scan_ids to free memory"""
        remove_scan_ids = []
        for scan_id, entry in self.sync_storage.items():
            if entry.get("status") not in ["closed", "aborted"]:
                continue
            if scan_id in self.scan_id_history:
                continue
            remove_scan_ids.append(scan_id)

        for scan_id in remove_scan_ids:
            for storage in [
                "sync_storage",
                "monitored_devices",
                "baseline_devices",
                "scan_motors",
                "readout_priority",
            ]:
                try:
                    getattr(self, storage).pop(scan_id)
                except KeyError:
                    logger.warning(f"Failed to remove {scan_id} from {storage}.")
            # self.bluesky_emitter.cleanup_storage(scan_id)
            self.run_emitter("on_cleanup", scan_id)
            self.storage_initialized.remove(scan_id)

    def _send_scan_point(self, scan_id, point_id) -> None:
        logger.info(f"Sending point {point_id} for scan_id {scan_id}.")
        logger.debug(f"{point_id}, {self.sync_storage[scan_id][point_id]}")

        self.run_emitter("on_scan_point_emit", scan_id, point_id)

        if point_id not in self.sync_storage[scan_id]["sent"]:
            self.sync_storage[scan_id]["sent"].add(point_id)
        else:
            logger.warning(f"Resubmitting existing point_id {point_id} for scan_id {scan_id}")

    def shutdown(self):
        self.device_manager.shutdown()
        self.connector.shutdown()
        self.executor.shutdown()
        for emi in self._emitter:
            emi.shutdown()



"""
This module provides the BECService class, which is the base class for all BEC services.
"""

from __future__ import annotations

import getpass
import socket
import threading
import time
import uuid
from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING, Any

import psutil
from rich.console import Console
from rich.table import Table

from bec_lib import messages
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from bec_lib.messages import BECStatus
from bec_lib.service_config import ServiceConfig

if TYPE_CHECKING:
    from bec_lib.connector import ConnectorBase


logger = bec_logger.logger

SERVICE_CONFIG = None


class BECService:
    def __init__(
        self,
        config: str | ServiceConfig,
        connector_cls: ConnectorBase,
        unique_service=False,
        wait_for_server=False,
    ) -> None:
        super().__init__()
        self._import_config(config)
        self._connector_cls = connector_cls
        self.connector = connector_cls(self.bootstrap_server)
        self._unique_service = unique_service
        self.wait_for_server = wait_for_server
        self.__service_id = str(uuid.uuid4())
        self._user = getpass.getuser()
        self._hostname = socket.gethostname()
        self._service_info_thread = None
        self._service_info_event = threading.Event()
        self._metrics_emitter_thread = None
        self._metrics_emitter_event = threading.Event()
        self._services_info = {}
        self._initialize_logger()
        self._check_services()
        self._status = BECStatus.BUSY
        self._start_update_service_info()
        self._start_metrics_emitter()
        self._wait_for_server()

    @property
    def _service_name(self):
        return (
            self.__class__.__name__
            if self._unique_service
            else f"{self.__class__.__name__}/{self._service_id}"
        )

    @property
    def _service_id(self):
        return self.__service_id

    def _import_config(self, config: str | ServiceConfig) -> None:
        if isinstance(config, str):
            self._service_config = ServiceConfig(config_path=config)
        elif isinstance(config, ServiceConfig):
            self._service_config = config
        else:
            raise TypeError("config must be of type str or ServiceConfig")
        global SERVICE_CONFIG
        SERVICE_CONFIG = self._service_config
        self.bootstrap_server = self._service_config.redis

    def _check_services(self, timeout_time=8, sleep_time=0.5) -> None:
        if not self._unique_service:
            return
        elapsed_time = 0
        while self._run_service_check(timeout_time, elapsed_time):
            elapsed_time += sleep_time
            time.sleep(sleep_time)

    def _run_service_check(self, timeout_time: float, elapsed_time: float) -> bool:
        self._update_existing_services()
        try:
            for service_name, msg in self._services_info.items():
                if service_name == self.__class__.__name__:
                    raise RuntimeError(
                        f"Another instance of {self.__class__.__name__} launched by user"
                        f" {msg.content['info']['user']} is already running on"
                        f" {msg.content['info']['hostname']}"
                    )
            return False
        except RuntimeError as service_error:
            if elapsed_time > timeout_time:
                raise RuntimeError from service_error
        return True

    def _initialize_logger(self) -> None:
        bec_logger.configure(
            self.bootstrap_server,
            self._connector_cls,
            service_name=self.__class__.__name__,
            service_config=self._service_config.config["service_config"],
        )

    def _update_existing_services(self):
        service_keys = self.connector.keys(MessageEndpoints.service_status("*").endpoint)
        if not service_keys:
            return
        services = [service.decode().split(":", maxsplit=1)[0] for service in service_keys]
        msgs = [self.connector.get(service) for service in services]
        self._services_info = {msg.content["name"]: msg for msg in msgs if msg is not None}

    def _update_service_info(self):
        while not self._service_info_event.is_set():
            logger.trace("Updating service info")
            try:
                self._send_service_status()
            except Exception:
                # exception is not explicitely specified,
                # because it depends on the underlying connector
                pass
            self._service_info_event.wait(timeout=3)

    def _send_service_status(self):
        self.connector.set_and_publish(
            topic=MessageEndpoints.service_status(self._service_id),
            msg=messages.StatusMessage(
                name=self._service_name,
                status=self.status,
                info={"user": self._user, "hostname": self._hostname, "timestamp": time.time()},
            ),
            expire=6,
        )

    @property
    def status(self) -> BECStatus:
        """get the current BECService status"""
        return self._status

    @status.setter
    def status(self, val: BECStatus):
        self._status = val
        self._send_service_status()

    def _start_update_service_info(self):
        self._service_info_thread = threading.Thread(
            target=self._update_service_info, daemon=True, name="update_service_info"
        )
        self._service_info_thread.start()

    def _start_metrics_emitter(self):
        self._metrics_emitter_thread = threading.Thread(
            target=self._get_metrics, daemon=True, name="metrics_emitter"
        )
        self._metrics_emitter_thread.start()

    def _get_metrics(self):
        proc = psutil.Process()
        while not self._metrics_emitter_event.is_set():
            res = proc.as_dict(
                attrs=[
                    "name",
                    "num_threads",
                    "pid",
                    "cpu_percent",
                    "memory_info",
                    "cmdline",
                    "cpu_times",
                    "create_time",
                    "memory_percent",
                ]
            )

            data = asdict(
                ServiceMetric(
                    process_name=res["name"],
                    username=self._user,
                    hostname=self._hostname,
                    cpu_percent=res["cpu_percent"],
                    cpu_times=res["cpu_times"].user,
                    cmdline=res["cmdline"],
                    num_threads=res["num_threads"],
                    pid=res["pid"],
                    memory_in_mb=res["memory_info"].rss / 1024 / 1024,
                    memory_used_percent=res["memory_percent"],
                    create_time=res["create_time"],
                )
            )
            msg = messages.ServiceMetricMessage(name=self.__class__.__name__, metrics=data)
            try:
                self.connector.send(MessageEndpoints.metrics(self._service_id), msg)
            except Exception:
                # exception is not explicitely specified,
                # because it depends on the underlying connector
                pass
            self._metrics_emitter_event.wait(timeout=1)

    def set_global_var(self, name: str, val: Any) -> None:
        """Set a global variable through Redis

        Args:
            name (str): Name of the variable
            val (Any): Value of the variable

        """
        self.connector.set(MessageEndpoints.global_vars(name), messages.VariableMessage(value=val))

    def get_global_var(self, name: str) -> Any:
        """Get a global variable from Redis

        Args:
            name (str): Name of the variable

        Returns:
            Any: Value of the variable
        """
        msg = self.connector.get(MessageEndpoints.global_vars(name))
        if msg:
            return msg.content.get("value")
        return None

    def delete_global_var(self, name: str) -> None:
        """Delete a global variable from Redis

        Args:
            name (str): Name of the variable

        """
        self.connector.delete(MessageEndpoints.global_vars(name))

    def global_vars(self) -> str:
        """Get all available global variables"""
        # sadly, this cannot be a property as it causes side effects with IPython's tab completion
        available_keys = self.connector.keys(MessageEndpoints.global_vars("*"))

        def get_endpoint_from_topic(topic: str) -> str:
            return topic.decode().split(MessageEndpoints.global_vars(""))[-1]

        endpoints = [get_endpoint_from_topic(k) for k in available_keys]

        console = Console()
        table = Table(title="Global variables")
        table.add_column("Variable", justify="center")
        table.add_column("Content", justify="center")
        for endpoint in endpoints:
            var = str(self.get_global_var(endpoint))
            if len(var) > 40:
                var = var[0:20] + "..., " + var[-20:]
            table.add_row(endpoint, var)
        with console.capture() as capture:
            console.print(table)
        out = capture.get()
        logger.info(out)
        print(out)

    def shutdown(self):
        """shutdown the BECService"""
        try:
            self.connector.shutdown()
            self._service_info_event.set()
            if self._service_info_thread:
                self._service_info_thread.join()
            self._metrics_emitter_event.set()
            if self._metrics_emitter_thread:
                self._metrics_emitter_thread.join()
        except AttributeError:
            print("Failed to shutdown BECService.")

    @property
    def service_status(self):
        """get the status of active services"""
        self._update_existing_services()
        return self._services_info

    def wait_for_service(self, name, status=BECStatus.RUNNING):
        logger.info(f"Waiting for {name}.")
        while True:
            service_status_msg = self.service_status.get(name)
            if service_status_msg is not None:
                service_status = BECStatus(service_status_msg.content["status"])
                if service_status == status:
                    break
            time.sleep(0.05)
        logger.success(f"{name} is running.")

    def _wait_for_server(self):
        if not self.wait_for_server:
            return
        try:
            self.wait_for_service("ScanServer", BECStatus.RUNNING)
            self.wait_for_service("ScanBundler", BECStatus.RUNNING)
            self.wait_for_service("DeviceServer", BECStatus.RUNNING)
            self.wait_for_service("SciHub", BECStatus.RUNNING)
            logger.success("All BEC services are running.")
        except KeyboardInterrupt:
            logger.warning("KeyboardInterrupt received. Stopped waiting for BEC services.")


@dataclass
class ServiceMetric:
    """Container for keeping performance metrics."""

    pid: int
    cmdline: list
    process_name: str
    username: str
    hostname: str
    cpu_percent: float
    cpu_times: dict
    num_threads: int
    memory_in_mb: float
    memory_used_percent: float
    create_time: float



from bec_ipython_client.prettytable import PrettyTable


def test_get_header():
    header = ["header1", "header2", "header3"]
    pt = PrettyTable(header)
    assert pt.get_header() == "|      header1     |      header2     |      header3     |"


def test_get_row():
    header = ["header1", "header2", "header3"]
    pt = PrettyTable(header)
    row = pt.get_row("row1", "row2", "row3")
    assert row == "|       row1       |       row2       |       row3       |"



from __future__ import annotations

import abc
import threading
import time
import traceback
from collections.abc import Callable
from typing import TYPE_CHECKING

from bec_lib import bec_logger, messages
from bec_lib.request_items import RequestItem

if TYPE_CHECKING:
    from bec_lib.client import BECClient

logger = bec_logger.logger


class ScanRequestError(Exception):
    """Error raised when a scan request is rejected"""


def set_event_delayed(event: threading.Event, delay: int) -> None:
    """Set event with a delay

    Args:
        event (threading.Event): event that should be set
        delay (int): delay time in seconds

    """

    def call_set():
        time.sleep(delay)
        event.set()

    thread = threading.Thread(target=call_set, daemon=True)
    thread.start()


def check_alarms(bec):
    """check for alarms and raise them if needed"""
    bec.alarm_handler.raise_alarms()


class LiveUpdatesBase(abc.ABC):
    def __init__(
        self,
        bec: BECClient,
        report_instruction: dict = None,
        request: messages.ScanQueueMessage = None,
        callbacks: list[Callable] = None,
    ) -> None:
        """Base class for live updates

        Args:
            bec (BECClient): BECClient instance
            report_instruction (dict, optional): report instruction. Defaults to None.
            request (messages.ScanQueueMessage, optional): scan queue request. Defaults to None.
            callbacks (list[Callable], optional): list of callback functions. Defaults to None.
        """
        self.bec = bec
        self.request = request
        self.RID = request.metadata["RID"]
        self.scan_queue_request = None
        self.report_instruction = report_instruction
        if callbacks is None:
            self.callbacks = []
        self.callbacks = callbacks if isinstance(callbacks, list) else [callbacks]

    def wait_for_request_acceptance(self):
        scan_request = ScanRequestMixin(self.bec, self.RID)
        scan_request.wait()
        self.scan_queue_request = scan_request.scan_queue_request

    @abc.abstractmethod
    def run(self):
        pass

    def emit_point(self, data: dict, metadata: dict = None):
        for cb in self.callbacks:
            if not cb:
                continue
            try:
                cb(data, metadata=metadata)
            except Exception:
                content = traceback.format_exc()
                logger.warning(f"Failed to run callback function: {content}")

    def _format_client_msg(self, msg: messages.ClientInfoMessage) -> str:
        """Pop messages from the client message handler.

        Args:
            messages (list, optional): List of messages to be popped. Defaults to None.
        """
        source = msg.source if msg.source else ""
        rtr = f"Client info ({source}) : {msg.message}"
        return rtr

    def _print_client_msgs_asap(self):
        """Print client messages flagged as show_asap"""
        # pylint: disable=protected-access
        if not self.scan_queue_request.client_messages_asap:
            return
        # pylint: disable=protected-access
        while len(self.scan_queue_request.client_messages_asap) > 0:
            msg = self.scan_queue_request.client_messages_asap.pop(0)
            print(self._format_client_msg(msg))

    # TODO #286
    # def _print_client_msgs_all(self):
    #     """Print summary of client messages"""
    #     # pylint: disable=protected-access
    #     if not self.scan_queue_request.client_messages:
    #         return
    #     print("------------------------")
    #     print("Summary of client messages")
    #     print("------------------------")
    #     # pylint: disable=protected-access
    #     while len(self.scan_queue_request.client_messages) > 0:
    #         msg = self.scan_queue_request.client_messages.pop(0)
    #         print(self._format_client_msg(msg))
    #     print("------------------------")


class ScanRequestMixin:
    def __init__(self, bec: BECClient, RID: str) -> None:
        """Mixin to handle scan request acceptance

        Args:
            bec (BECClient): BECClient instance
            RID (str): request ID
        """
        self.bec = bec
        self.request_storage = self.bec.queue.request_storage
        self.RID = RID
        self.scan_queue_request = None

    def _wait_for_scan_request(self) -> RequestItem:
        """wait for scan queuest

        Returns:
            RequestItem: scan queue request
        """
        logger.debug("Waiting for request ID")
        start = time.time()
        while self.request_storage.find_request_by_ID(self.RID) is None:
            time.sleep(0.1)
            check_alarms(self.bec)
        logger.debug(f"Waiting for request ID finished after {time.time()-start} s.")
        return self.request_storage.find_request_by_ID(self.RID)

    def _wait_for_scan_request_decision(self):
        """wait for a scan queuest decision"""
        logger.debug("Waiting for decision")
        start = time.time()
        while self.scan_queue_request.decision_pending:
            time.sleep(0.1)
            check_alarms(self.bec)
        logger.debug(f"Waiting for decision finished after {time.time()-start} s.")

    def wait(self):
        """wait for the request acceptance"""
        self.scan_queue_request = self._wait_for_scan_request()

        self._wait_for_scan_request_decision()
        check_alarms(self.bec)

        while self.scan_queue_request.accepted is None:
            time.sleep(0.1)
            check_alarms(self.bec)

        if not self.scan_queue_request.accepted[0]:
            raise ScanRequestError(
                "Scan was rejected by the server:"
                f" {self.scan_queue_request.response.content.get('message')}"
            )

        while self.scan_queue_request.queue is None:
            time.sleep(0.1)
            check_alarms(self.bec)



from __future__ import annotations

from typing import TYPE_CHECKING

from IPython.core.magic import Magics, line_magic, magics_class

if TYPE_CHECKING:
    from bec_ipython_client import BECClient


@magics_class
class BECMagics(Magics):
    def __init__(self, shell, client: BECClient):
        super(BECMagics, self).__init__(shell)
        self.client = client

    @line_magic
    def abort(self, line):
        "Request a scan abortion"
        return self.client.queue.request_scan_abortion()

    @line_magic
    def reset(self, line):
        "Request a scan queue reset"
        return self.client.queue.request_queue_reset()

    @line_magic
    def resume(self, line):
        "Resume the scan"
        self.client.queue.request_scan_continuation()
        return self.client.live_updates.continue_request()

    @line_magic
    def pause(self, line):
        "Request a scan pause"
        return self.client.queue.request_scan_interruption(deferred_pause=False)

    @line_magic
    def deferred_pause(self, line):
        "Request a deferred pause"
        return self.client.queue.request_scan_interruption(deferred_pause=True)

    @line_magic
    def restart(self, line):
        "Request a scan restart"
        old_req_ids = self.client.queue.scan_storage.current_scan.queue.requestIDs
        request = self.client.queue.request_storage.find_request_by_ID(old_req_ids[0]).request
        requestID = self.client.queue.request_scan_restart()
        request.metadata["RID"] = requestID
        hide_report = request.metadata.get("hide_report", False)
        # pylint: disable=protected-access
        scan_report = self.client.scans._available_scans["fermat_scan"]._get_scan_report_type(
            hide_report
        )
        return self.client.live_updates.process_request(request, scan_report, [])

    @line_magic
    def halt(self, line):
        "Request a scan halt, i.e. abort without cleanup."
        return self.client.queue.request_scan_halt()



import time

import pytest

from bec_server.scihub.repeated_timer import RepeatedTimer


class Counter:
    def __init__(self):
        self.count = 0

    def increment(self):
        self.count += 1


@pytest.fixture
def counter():
    return Counter()


@pytest.fixture
def timer(counter):
    return RepeatedTimer(0.1, counter.increment)


def test_start_and_stop(timer, counter):
    timer.start()
    time.sleep(2)
    timer.stop()

    # accept some variance in the count due to timing issues
    assert 18 <= counter.count <= 20



import os
import subprocess


def subprocess_start(bec_path: str, config_path: str, services: dict):
    processes = []
    for ii, service_info in enumerate(services.items()):
        service, service_config = service_info
        service_path = service_config["path"].substitute(base_path=bec_path)
        # service_config adds a subdirectory to each path, here we do not want the subdirectory
        cwd = os.path.abspath(os.path.join(service_path, ".."))
        if config_path:
            processes.append(
                subprocess.Popen((service_config["command"], "--config", config_path), cwd=cwd)
            )
        else:
            processes.append(subprocess.Popen((service_config["command"],), cwd=cwd))
    return processes


def subprocess_stop():
    # do nothing for now... would require pid files or something to keep track
    # of the started processes
    ...



import argparse
import os
import sys
from importlib.metadata import version
from typing import Iterable, Literal

import IPython
from IPython.terminal.ipapp import TerminalIPythonApp
from IPython.terminal.prompts import Prompts, Token

from bec_ipython_client.beamline_mixin import BeamlineMixin
from bec_ipython_client.bec_magics import BECMagics
from bec_ipython_client.callbacks.ipython_live_updates import IPythonLiveUpdates
from bec_ipython_client.signals import ScanInterruption, SigintHandler
from bec_lib import ServiceConfig, bec_logger, plugin_helper
from bec_lib.alarm_handler import AlarmBase
from bec_lib.callback_handler import EventType
from bec_lib.client import BECClient
from bec_lib.connector import ConnectorBase

logger = bec_logger.logger


class BECIPythonClient:
    def __init__(
        self,
        config: ServiceConfig = None,
        connector_cls: ConnectorBase = None,
        wait_for_server=True,
        forced=False,
    ) -> None:
        self._client = BECClient(config, connector_cls, wait_for_server, forced, parent=self)
        self._ip = IPython.get_ipython()
        self.started = False
        self._sighandler = None
        self._beamline_mixin = None
        self._exit_event = None
        self._exit_handler_thread = None
        self.live_updates = None
        self.gui = None
        self._client.callbacks.register(
            event_type=EventType.NAMESPACE_UPDATE, callback=self._update_namespace_callback
        )
        self._client.load_high_level_interface("bec_hli")

    def __getattr__(self, name):
        return getattr(self._client, name)

    def __dir__(self) -> Iterable[str]:
        return dir(self._client) + dir(self.__class__)

    def __str__(self) -> str:
        return "BECIPythonClient\n\nTo get a list of available commands, type `bec.show_all_commands()`"

    def start(self):
        """start the client"""
        if self.started:
            return
        self._client.start()
        self._sighandler = SigintHandler(self)
        self._beamline_mixin = BeamlineMixin()
        self.live_updates = IPythonLiveUpdates(self)
        self._configure_ipython()
        self.started = True

    def bl_show_all(self):
        self._beamline_mixin.bl_show_all()

    def _set_ipython_prompt_scan_number(self, scan_number: int):
        if self._ip:
            self._ip.prompts.scan_number = scan_number + 1

    def _configure_ipython(self):

        if self._ip is None:
            return

        self._ip.prompts = BECClientPrompt(ip=self._ip, client=self._client, username="demo")
        self._load_magics()
        self._ip.events.register("post_run_cell", log_console)
        self._ip.set_custom_exc((Exception,), _ip_exception_handler)  # register your handler
        # represent objects using __str__, if overwritten, otherwise use __repr__
        self._ip.display_formatter.formatters["text/plain"].for_type(
            object,
            lambda o, p, cycle: o.__str__ is object.__str__ and p.text(repr(o)) or p.text(str(o)),
        )

    def _update_namespace_callback(self, action: Literal["add", "remove"], ns_objects: dict):
        """Callback to update the global namespace of ipython.

        Within BEC, the callback is triggered by the CallbackHandler when
        the namespace changes, e.g. load_high_level_interface, load_user_script, etc.

        Args:
            action (Literal["add", "remove"]): action to perform
            ns_objects (dict): objects to add or remove

        """
        if self._ip is None:
            return
        if action == "add":
            for name, obj in ns_objects.items():
                self._ip.user_global_ns[name] = obj
        elif action == "remove":
            for name, obj in ns_objects.items():
                self._ip.user_global_ns.pop(name)

    def _set_error(self):
        if self._ip is not None:
            self._ip.prompts.status = 0

    def _set_busy(self):
        if self._ip is not None:
            self._ip.prompts.status = 1

    def _set_idle(self):
        if self._ip is not None:
            self._ip.prompts.status = 2

    def _load_magics(self):
        magics = BECMagics(self._ip, self._client)
        self._ip.register_magics(magics)

    def shutdown(self):
        """shutdown the client and all its components"""
        if self.gui:
            self.gui.close()

        self._client.shutdown()


def _ip_exception_handler(self, etype, evalue, tb, tb_offset=None):
    if issubclass(etype, (AlarmBase, ScanInterruption)):
        print(f"\x1b[31m BEC alarm:\x1b[0m: {evalue}")
        return
    self.showtraceback((etype, evalue, tb), tb_offset=None)  # standard IPython's printout


class BECClientPrompt(Prompts):
    def __init__(self, ip, username, client, status=0):
        self._username = username
        self.client = client
        self.status = status
        super().__init__(ip)

    def in_prompt_tokens(self, cli=None):
        if self.status == 0:
            status_led = Token.OutPromptNum
        elif self.status == 1:
            status_led = Token.PromptNum
        else:
            status_led = Token.Prompt
        try:
            next_scan_number = str(self.client.queue.next_scan_number)
        except Exception:
            next_scan_number = "?"
        return [
            (status_led, "\u2022"),
            (Token.Prompt, " " + self.username),
            (Token.Prompt, " ["),
            (Token.PromptNum, str(self.shell.execution_count)),
            (Token.Prompt, "/"),
            (Token.PromptNum, next_scan_number),
            (Token.Prompt, "] "),
            (Token.Prompt, "❯❯ "),
        ]

    @property
    def username(self):
        """current username"""
        return self._username

    @username.setter
    def username(self, value):
        self._username = value


def log_console(execution_info):
    """log the console input"""
    logger.log("CONSOLE_LOG", f"{execution_info.info.raw_cell}")


# pylint: disable=wrong-import-position
# pylint: disable=protected-access
# pylint: disable=unused-import
# pylint: disable=ungrouped-imports

main_dict = {}

sys.modules["bec_ipython_client.main"] = sys.modules[
    __name__
]  # properly register module when file is executed directly, like in test


def main():
    """
    Main function to start the BEC IPython client.
    """

    available_plugins = plugin_helper.get_ipython_client_startup_plugins(state="pre")

    parser = argparse.ArgumentParser(
        prog="BEC IPython client", description="BEC command line interface"
    )
    parser.add_argument("--version", action="store_true", default=False)
    parser.add_argument("--nogui", action="store_true", default=False)
    parser.add_argument("--config", action="store", default=None)
    parser.add_argument("--dont-wait-for-server", action="store_true", default=False)
    parser.add_argument("--post-startup-file", action="store", default=None)

    for plugin in available_plugins.values():
        if hasattr(plugin["module"], "extend_command_line_args"):
            plugin["module"].extend_command_line_args(parser)

    args, left_args = parser.parse_known_args()

    # remove already parsed args from command line args
    sys.argv = sys.argv[:1] + left_args

    if args.version:
        print(f"BEC IPython client: {version('bec_ipython_client')}")
        sys.exit(0)

    config_file = args.config
    if config_file:
        if not os.path.isfile(config_file):
            raise FileNotFoundError("Config file not found.")
        print("Using config file: ", config_file)
        config = ServiceConfig(config_file)

    if available_plugins and "config" not in locals():
        # check if pre-startup.py script exists
        for plugin in available_plugins.values():
            if hasattr(plugin["module"], "get_config"):
                config = plugin["module"].get_config()
                break

    # check if config was defined in pre-startup.py
    if "config" not in locals():
        config = ServiceConfig()

    main_dict["config"] = config
    main_dict["args"] = args
    main_dict["wait_for_server"] = not args.dont_wait_for_server
    main_dict["startup_file"] = args.post_startup_file

    app = TerminalIPythonApp()
    app.interact = True
    app.initialize(argv=["-i", os.path.join(os.path.dirname(__file__), "bec_startup.py")])

    try:
        app.start()
    finally:
        if "bec" in main_dict:
            # bec object is inserted into main_dict by bec_startup
            main_dict["bec"].shutdown()


if __name__ == "__main__":
    main()



import os

import pytest

from bec_lib.bec_yaml_loader import yaml_load


@pytest.fixture
def test_file1():
    return "eiger:\n  readoutPriority: monitored\n  deviceClass: ophyd_devices.SimCamera\n  deviceConfig:\n    device_access: true\n  deviceTags:\n    - detector\n  enabled: true\n  readOnly: false\n  softwareTrigger: true"


@pytest.fixture
def test_file2():
    return "samx:\n  readoutPriority: monitored\n  deviceClass: ophyd_devices.SimCamera\n  deviceConfig:\n    device_access: true\n  deviceTags:\n    - detector\n  enabled: true\n  readOnly: false\n  softwareTrigger: true"


@pytest.fixture
def test_file3():
    return "samy:\n  readoutPriority: monitored\n  deviceClass: ophyd_devices.SimCamera\n  deviceConfig:\n    device_access: true\n  deviceTags:\n    - detector\n  enabled: true\n  readOnly: false\n  softwareTrigger: true"


def _remove_files(files):
    for file in files:
        if os.path.exists(file):
            os.remove(file)


def test_load_yaml_without_include(test_file1):
    # sastt:
    #   - !include /Users/wakonig_k/software/work/csaxs-bec/csaxs_bec/device_configs/bec_device_config_sastt.yaml
    output_file_1 = test_file1
    with open("test_file_1.yaml", "w", encoding="utf-8") as file:
        file.write(output_file_1)
    try:
        out = yaml_load("test_file_1.yaml")
    finally:
        _remove_files(["test_file_1.yaml"])

    assert "eiger" in out
    assert len(out) == 1


def test_load_yaml_single_include(test_file1, test_file2):
    # sastt:
    #   - !include /Users/wakonig_k/software/work/csaxs-bec/csaxs_bec/device_configs/bec_device_config_sastt.yaml
    include_str = "sastt: !include ./test_file2.yaml"
    output_file_1 = test_file1 + "\n" + include_str
    output_file_2 = test_file2
    with open("test_file_1.yaml", "w", encoding="utf-8") as file:
        file.write(output_file_1)
    with open("test_file2.yaml", "w", encoding="utf-8") as file:
        file.write(output_file_2)
    try:
        out = yaml_load("test_file_1.yaml")
    finally:
        _remove_files(["test_file_1.yaml", "test_file2.yaml"])

    assert "samx" in out
    assert "eiger" in out
    assert len(out) == 2


def test_load_yaml_single_include_with_conflict(capfd, test_file1):
    # sastt:
    #   - !include /Users/wakonig_k/software/work/csaxs-bec/csaxs_bec/device_configs/bec_device_config_sastt.yaml
    include_str = "sastt: !include ./test_file2.yaml"
    output_file_1 = test_file1 + "\n" + include_str
    output_file_2 = test_file1
    output_file_1.replace(
        "deviceClass: ophyd_devices.SimCamera", "deviceClass: ophyd_devices.Eiger"
    )
    with open("test_file_1.yaml", "w", encoding="utf-8") as file:
        file.write(output_file_1)
    with open("test_file2.yaml", "w", encoding="utf-8") as file:
        file.write(output_file_2)
    # capture stdout
    try:
        out = yaml_load("test_file_1.yaml")
    finally:
        _remove_files(["test_file_1.yaml", "test_file2.yaml"])

    assert "eiger" in out
    assert len(out) == 1
    assert out["eiger"]["deviceClass"] == "ophyd_devices.SimCamera"
    out, _ = capfd.readouterr()
    assert (
        "Warning: Multiple definitions for key eiger. Using the one from ./test_file2.yaml." in out
    )


def test_load_yaml_multi_include(test_file1, test_file2, test_file3):
    # sastt:
    #   - !include /Users/wakonig_k/software/work/csaxs-bec/csaxs_bec/device_configs/bec_device_config_sastt.yaml
    include_str = "sastt:\n  - !include ./test_file2.yaml\n  - !include ./test_file3.yaml"
    output_file_1 = test_file1 + "\n" + include_str
    output_file_2 = test_file2
    output_file_3 = test_file3
    with open("test_file_1.yaml", "w", encoding="utf-8") as file:
        file.write(output_file_1)
    with open("test_file2.yaml", "w", encoding="utf-8") as file:
        file.write(output_file_2)
    with open("test_file3.yaml", "w", encoding="utf-8") as file:
        file.write(output_file_3)
    try:
        out = yaml_load("test_file_1.yaml")
    finally:
        _remove_files(["test_file_1.yaml", "test_file2.yaml", "test_file3.yaml"])

    assert "samx" in out
    assert "samy" in out
    assert "eiger" in out
    assert len(out) == 3



from bec_lib.devicemanager import DeviceContainer
from bec_lib.tests.utils import ConnectorMock


# pylint: disable=missing-function-docstring
# pylint: disable=protected-access
class DeviceMock:
    def __init__(self, name: str, value=None):
        self.name = name
        self.read_buffer = {self.name: {"value": value}} if value is not None else None
        self._config = {"deviceConfig": {"limits": [-50, 50]}, "userParameter": None}
        self._read_only = False
        self._enabled = True

    def read(self):
        return self.read_buffer

    def readback(self):
        return self.read_buffer

    @property
    def obj(self):
        return self

    @property
    def limits(self):
        return self._config["deviceConfig"]["limits"]

    @property
    def read_only(self) -> bool:
        return self._read_only

    @read_only.setter
    def read_only(self, val: bool):
        self._read_only = val

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, val: bool):
        self._enabled = val

    @property
    def user_parameter(self):
        return self._config["userParameter"]


class DMMock:
    devices = DeviceContainer()
    connector = ConnectorMock()

    def add_device(self, name, value=None):
        self.devices[name] = DeviceMock(name, value=value)



import time

from bec_ipython_client.progressbar import ScanProgressBar
from bec_lib import MessageEndpoints, bec_logger

from .live_table import LiveUpdatesTable

logger = bec_logger.logger


class LiveUpdatesScanProgress(LiveUpdatesTable):
    """Live updates for scans using a progress bar based on the progress of one or more devices"""

    REPORT_TYPE = "scan_progress"

    def _run_update(self, device_names: str):
        """Run the update loop for the progress bar.

        Args:
            device_names (str): The name of the device to monitor.
        """
        with ScanProgressBar(
            scan_number=self.scan_item.scan_number, clear_on_exit=False
        ) as progressbar:
            while True:
                if self._update_progressbar(progressbar, device_names):
                    break
                self._print_client_msgs_asap()

        # TODO #286
        # self._print_client_msgs_all()

    def _update_progressbar(self, progressbar: ScanProgressBar, device_names: str) -> bool:
        """Update the progressbar based on the device status message

        Args:
            progressbar (ScanProgressBar): The progressbar to update.
            device_names (str): The name of the device to monitor.
        Returns:
            bool: True if the scan is finished.
        """
        self.check_alarms()
        status = self.bec.connector.get(MessageEndpoints.device_progress(device_names[0]))
        if not status:
            logger.debug("waiting for new data point")
            time.sleep(0.1)
            return False
        if status.metadata.get("scan_id") != self.scan_item.scan_id:
            logger.debug("waiting for new data point")
            time.sleep(0.1)
            return False

        point_id = status.content.get("value")
        if point_id is None:
            logger.debug("waiting for new data point")
            time.sleep(0.1)
            return False

        max_value = status.content.get("max_value")
        if max_value and max_value != progressbar.max_points:
            progressbar.max_points = max_value

        progressbar.update(point_id)
        # process sync callbacks
        self.bec.callbacks.poll()
        self.scan_item.poll_callbacks()

        done = status.content.get("done")
        if point_id == max_value or done:
            return True
        return False



from __future__ import annotations

import time
from collections.abc import Callable
from typing import TYPE_CHECKING

from bec_ipython_client.prettytable import PrettyTable
from bec_ipython_client.progressbar import ScanProgressBar
from bec_lib import bec_logger, messages

from .utils import LiveUpdatesBase, check_alarms

if TYPE_CHECKING:
    from bec_lib.client import BECClient

logger = bec_logger.logger


def sort_devices(devices, scan_devices) -> list:
    """sort the devices to ensure that the table starts with scan motors"""
    for scan_dev in list(scan_devices)[::-1]:
        devices.remove(scan_dev)
        devices.insert(0, scan_dev)
    return devices


class LiveUpdatesTable(LiveUpdatesBase):
    """Live updates for scans using a table and a scan progess bar.

    Args:
        bec (BECClient): client instance
        request (messages.ScanQueueMessage): The scan request that should be monitored

    Raises:
        TimeoutError: Raised if no queue item is added before reaching a predefined timeout.
        RuntimeError: Raised if more points than requested are returned.
        ScanRequestError: Raised if the scan was rejected by the server.
    """

    MAX_DEVICES = 10
    REPORT_TYPE = "table_wait"

    def __init__(
        self,
        bec: BECClient,
        report_instruction: dict = None,
        request: messages.ScanQueueMessage = None,
        callbacks: list[Callable] = None,
        print_table_data=None,
    ) -> None:
        super().__init__(
            bec, report_instruction=report_instruction, request=request, callbacks=callbacks
        )
        self.scan_queue_request = None
        self.scan_item = None
        self.dev_values = None
        self.point_data = None
        self.point_id = 0
        self.table = None
        self.__print_table_data = None
        self._print_table_data = (
            print_table_data if print_table_data is not None else self.REPORT_TYPE == "table_wait"
        )

    def wait_for_scan_to_start(self):
        """wait until the scan starts"""
        while True:
            queue_pos = self.scan_item.queue.queue_position
            self.check_alarms()
            if self.scan_item.status == "closed":
                break
            if queue_pos is None:
                logger.debug(f"Could not find queue entry for scan_id {self.scan_item.scan_id}")
                continue
            if queue_pos == 0:
                break
            print(
                f"Scan is enqueued and is waiting for execution. Current position in queue: {queue_pos + 1}.",
                end="\r",
                flush=True,
            )
            time.sleep(0.1)
        while not self.scan_item.scan_number:
            time.sleep(0.05)

    def wait_for_scan_item_to_finish(self):
        """wait for scan completion"""
        while True:
            if self.scan_item.end_time:
                if self.scan_item.open_queue_group:
                    break
                if self.scan_item.queue.queue_position is None:
                    break
            self.check_alarms()
            time.sleep(0.1)

    def check_alarms(self):
        """check for alarms"""
        check_alarms(self.bec)

    def resume(self, request: messages.ScanQueueMessage, report_instruction: str, callbacks):
        """resume the scan after a pause"""
        super().__init__(
            self.bec, request=request, report_instruction=report_instruction, callbacks=callbacks
        )
        self.process_request()

    @property
    def devices(self):
        """get the devices for the callback"""
        if self.point_data.metadata["scan_type"] == "step":
            return self.get_devices_from_scan_data(self.scan_item.data[0])
        if self.point_data.metadata["scan_type"] == "fly":
            devices = list(self.point_data.content["data"].keys())
            if len(devices) > self.MAX_DEVICES:
                return devices[0 : self.MAX_DEVICES]
            return devices
        return None

    def get_devices_from_scan_data(self, data: messages.ScanMessage) -> list:
        """extract interesting devices from a scan request"""
        device_manager = self.bec.device_manager
        scan_devices = data.metadata.get("scan_report_devices")
        monitored_devices = device_manager.devices.monitored_devices(
            [device_manager.devices[dev] for dev in scan_devices]
        )
        # devices = [hint for dev in monitored_devices for hint in dev._hints]
        devices = [dev.name for dev in monitored_devices]
        devices = sort_devices(devices, scan_devices)
        if len(devices) > self.MAX_DEVICES:
            return devices[0 : self.MAX_DEVICES]
        return devices

    def _prepare_table(self) -> PrettyTable:
        """Prepares the custom table for the live updates."""
        header = self._get_header()
        max_len = max(len(head) for head in header)
        return PrettyTable(header, padding=max_len)

    def _get_header(self) -> list:
        """get the header for the table with up to self.MAX_DEVICES as entries

        Returns:
            list: header for the table
        """
        header = ["seq. num"]
        for dev in self.devices:
            if dev in self.bec.device_manager.devices:
                obj = self.bec.device_manager.devices[dev]
                header.extend(obj._hints)
            else:
                header.append(dev)
        return header

    def update_scan_item(self, timeout: float = 15):
        """get the current scan item"""
        start = time.time()
        while self.scan_queue_request.scan is None:
            self.check_alarms()
            time.sleep(0.1)
            if time.time() - start > timeout:
                raise TimeoutError("Could not find scan item.")
        self.scan_item = self.scan_queue_request.scan

    def core(self):
        """core function to run the live updates for the table"""
        req_ID = self.scan_queue_request.requestID
        while True:
            request_block = [
                req for req in self.scan_item.queue.request_blocks if req["RID"] == req_ID
            ][0]
            if not request_block["is_scan"]:
                break
            if request_block["report_instructions"]:
                break
            self.check_alarms()

        self._run_update(self.report_instruction[self.REPORT_TYPE])

    def _run_update(self, target_num_points: int):
        """run the update loop with the progress bar

        Args:
            target_num_points (int): number of points to be collected
        """
        with ScanProgressBar(
            scan_number=self.scan_item.scan_number, clear_on_exit=self._print_table_data
        ) as progressbar:
            while True:
                self.check_alarms()
                self.point_data = self.scan_item.data.get(self.point_id)
                if self.scan_item.num_points:
                    progressbar.max_points = self.scan_item.num_points
                    if target_num_points == 0:
                        target_num_points = self.scan_item.num_points

                progressbar.update(self.point_id)
                if self.point_data:
                    self.point_id += 1
                    self.print_table_data()
                    progressbar.update(self.point_id)

                    # process sync callbacks
                    self.bec.callbacks.poll()
                    self.scan_item.poll_callbacks()
                else:
                    logger.debug("waiting for new data point")
                    time.sleep(0.1)

                if not self.scan_item.num_points:
                    continue

                if self.point_id == target_num_points:
                    break
                if self.point_id > self.scan_item.num_points:
                    raise RuntimeError("Received more points than expected.")

    @property
    def _print_table_data(self) -> bool:
        """Checks if the table should be printed or not.

        Returns:
            bool: True/False
        """
        if not self.__print_table_data:
            return False
        try:
            # pylint: disable=protected-access
            # pylint: disable=undefined-variable
            if get_ipython().__class__.__name__ == "ZMQInteractiveShell":
                self.__print_table_data = False
                return False
        except Exception:
            pass
        return self.__print_table_data

    @_print_table_data.setter
    def _print_table_data(self, value: bool):
        """Set the print_table_data attribute.

        Args:
            value (bool): value to set
        """
        self.__print_table_data = value

    def print_table_data(self):
        """print the table data for the current point_id"""
        if not self._print_table_data:
            return

        if not self.table:
            self.dev_values = (len(self._get_header()) - 1) * [0]
            self.table = self._prepare_table()
            print(self.table.get_header_lines())

        if self.point_id % 100 == 0:
            print(self.table.get_header_lines())
        ind = 0
        for dev in self.devices:
            if dev in self.bec.device_manager.devices:
                obj = self.bec.device_manager.devices[dev]
                for hint in obj._hints:
                    signal = self.point_data.content["data"].get(dev, {}).get(hint)
                    if signal is None or signal.get("value") is None:
                        print_value = "N/A"
                    else:
                        try:
                            precision = getattr(obj, "precision")
                        except AttributeError:
                            precision = 2
                        value = signal.get("value")
                        if isinstance(value, (int, float)):
                            print_value = f"{value:.{precision}f}"
                        else:
                            print_value = str(value)
                    self.dev_values[ind] = print_value
                    ind += 1
            else:
                signal = self.point_data.content["data"].get(dev, {})
                value = signal.get("value")
                if value is not None:
                    if isinstance(value, (int, float)):
                        print_value = f"{value:.2f}"
                    else:
                        print_value = str(value)
                else:
                    print_value = "N/A"
                self.dev_values[ind] = print_value
                ind += 1
        print(self.table.get_row(str(self.point_id), *self.dev_values))
        # pylint: disable=protected-access
        self._print_client_msgs_asap()

    def close_table(self):
        """close the table and print the footer"""
        if not self.table:
            return
        elapsed_time = self.scan_item.end_time - self.scan_item.start_time
        print(
            self.table.get_footer(
                f"Scan {self.scan_item.scan_number} finished. Scan ID {self.scan_item.scan_id}. Elapsed time: {elapsed_time:.2f} s"
            )
        )
        # TODO #286
        # self._print_client_msgs_all()

    def process_request(self):
        """process the request and start the core loop for live updates"""
        if self.request.content["scan_type"] == "close_scan_def":
            self.wait_for_scan_item_to_finish()
            self.close_table()
            return

        self.wait_for_request_acceptance()
        self.update_scan_item(timeout=15)
        self.wait_for_scan_to_start()

        if self.table:
            self.table = None
        else:
            if self._print_table_data:
                print(f"\nStarting scan {self.scan_item.scan_number}.")

        self.core()

    def run(self):
        """run the live updates"""
        if self.request.content["scan_type"] == "open_scan_def":
            self.wait_for_request_acceptance()
            return
        self.process_request()
        self.wait_for_scan_item_to_finish()
        if self._print_table_data:
            self.close_table()



"""
RPC utility functions for the bec_lib package.
"""


class user_access:

    def __init__(self, meth):
        """
        Decorator to mark <device_class> methods to be accessible from the bec client

        Args:
            meth: method to be marked as accessible

        Examples:
            >>> @user_access
            >>> def my_method(self, *args, **kwargs):
            >>>     return fcn(self, *args, **kwargs)

        """
        self.meth = meth

    def __set_name__(self, owner, name):
        """Write registered method into the owner class USER_ACCESS list for the bec_ipython_client"""
        if not hasattr(owner, "USER_ACCESS"):
            owner.USER_ACCESS = []
        if name not in owner.USER_ACCESS:
            owner.USER_ACCESS.append(name)

        setattr(owner, name, self.meth)



from unittest import mock

import msgpack
import pytest

from bec_lib import MessageEndpoints
from bec_server.scan_bundler.bluesky_emitter import BlueskyEmitter


@pytest.fixture
def bls_emitter(scan_bundler_mock):
    emitter = BlueskyEmitter(scan_bundler_mock)
    yield emitter
    emitter.shutdown()


@pytest.mark.parametrize("scan_id", ["alskdj"])
def test_run_start_document(bls_emitter, scan_id):
    with mock.patch.object(bls_emitter.connector, "raw_send") as send:
        with mock.patch.object(bls_emitter, "send_descriptor_document") as send_descr:
            with mock.patch.object(
                bls_emitter, "_get_run_start_document", return_value={}
            ) as get_doc:
                bls_emitter.send_run_start_document(scan_id)
                get_doc.assert_called_once_with(scan_id)
                send.assert_called_once_with(
                    MessageEndpoints.bluesky_events(), msgpack.dumps(("start", {}))
                )
                send_descr.assert_called_once_with(scan_id)


def test_get_run_start_document(bls_emitter):
    sb = bls_emitter.scan_bundler
    scan_id = "lkajsdl"
    sb.sync_storage[scan_id] = {"info": {"queue_id": "jdklj", "scan_number": 5}}
    sb.scan_motors[scan_id] = [sb.device_manager.devices.samx, sb.device_manager.devices.samy]

    data = bls_emitter._get_run_start_document(scan_id)

    assert all(key in data for key in ["time", "uid", "scan_id", "queue_id", "scan_id", "motors"])
    assert data["motors"] == ("samx", "samy")
    assert data["scan_id"] == 5


def test_send_descriptor_document(bls_emitter):
    scan_id = "lkajsdl"
    bls_emitter.bluesky_metadata[scan_id] = {}
    with mock.patch.object(bls_emitter.connector, "raw_send") as send:
        with mock.patch.object(
            bls_emitter, "_get_descriptor_document", return_value={}
        ) as get_descr:
            bls_emitter.send_descriptor_document(scan_id)
            get_descr.assert_called_once_with(scan_id)
            send.assert_called_once_with(
                MessageEndpoints.bluesky_events(), msgpack.dumps(("descriptor", {}))
            )


def test_bls_cleanup_storage(bls_emitter):
    scan_id = "lkajsdl"
    bls_emitter.bluesky_metadata[scan_id] = {}

    bls_emitter.cleanup_storage(scan_id)
    assert scan_id not in bls_emitter.bluesky_metadata


def test_bls_on_cleanup(bls_emitter):
    scan_id = "lkajsdl"
    with mock.patch.object(bls_emitter, "cleanup_storage") as cleanup:
        bls_emitter.on_cleanup(scan_id)
        cleanup.assert_called_once_with(scan_id)


def test_bls_on_init(bls_emitter):
    scan_id = "lkajsdl"
    with mock.patch.object(bls_emitter, "send_run_start_document") as start:
        bls_emitter.on_init(scan_id)
        start.assert_called_once_with(scan_id)



import numpy as np


class PathOptimizerMixin:
    def get_radius(self, pos):
        return np.sqrt(np.abs(pos[:, 0]) ** 2 + np.abs(pos[:, 1]) ** 2)

    def optimize_corridor(self, positions, corridor_size=None, sort_axis=1, num_iterations=1):
        if len(positions[0]) < 2:
            return positions
        if corridor_size is None:
            dims = [
                abs(min(positions[:, 0]) - max(positions[:, 0])),
                abs(min(positions[:, 1]) - max(positions[:, 1])),
            ]
            density = np.sqrt(len(positions) / (dims[1] * dims[0]))
            corridor_size = 2 / density

        result = [np.inf, []]
        if num_iterations > 1:
            corridor_sizes = np.linspace(corridor_size / 2, corridor_size * 2, num_iterations)
        else:
            corridor_sizes = [corridor_size]
        for corridor_size_iter in corridor_sizes:
            index_sorted = []

            n_steps = int(
                np.round(
                    np.abs(max(positions[:, sort_axis]) - min(positions[:, sort_axis]))
                    / corridor_size_iter
                )
            )
            steps = list(
                np.floor(
                    np.linspace(min(positions[:, sort_axis]), max(positions[:, sort_axis]), n_steps)
                )
            )
            # print(steps, n_steps, positions)
            steps.append(np.inf)
            for step in range(n_steps):
                block = np.where(positions[:, sort_axis] >= steps[step])
                block = block[0][np.where(positions[:, sort_axis][block[0]] < steps[step + 1])]
                block = block[np.argsort(positions[:, int(not sort_axis)][block])]
                if step % 2 == 0:
                    block = block[::-1]
                index_sorted.append(block)
            path_length = self.get_path_length(positions[np.concatenate(index_sorted), :])
            if path_length < result[0]:
                result = [path_length, positions[np.concatenate(index_sorted), :]]
        return result[1]

    def optimize_shell(self, pos, offset, dr):
        max_rad_id = np.where(self.get_radius(pos) == np.max(self.get_radius(pos)))
        max_rad = self.get_radius(pos[max_rad_id])[0]

        sub_groups = []
        nsteps = int(np.floor(max_rad / dr) + int(bool(np.mod(max_rad, dr))))
        sub_rad_prev = -offset * dr
        for i in range(nsteps + 2):
            temp = np.where(self.get_radius(pos) < dr + sub_rad_prev)
            temp = temp[0][np.where(self.get_radius(pos[temp[0]]) >= sub_rad_prev)]
            temp_pos = pos[temp]
            angles = np.arctan2(temp_pos[:, 0], (temp_pos[:, 1]))
            pos_sort = temp_pos[np.argsort(angles)]
            sub_groups.append(pos_sort)
            sub_rad_prev += dr

        temp = np.where(self.get_radius(pos) >= sub_rad_prev)
        temp = temp[0][np.where(self.get_radius(pos[temp[0]]) >= sub_rad_prev)]
        temp_pos = pos[temp]
        angles = np.arctan2(temp_pos[:, 0], (temp_pos[:, 1]))
        pos_sort = temp_pos[np.argsort(angles)]
        sub_groups.append(pos_sort)

        posi = []
        for i in range(len(sub_groups)):
            for j in range(len(sub_groups[i])):
                posi.extend([sub_groups[i][j]])

        return np.asarray(posi)

    def get_path_length(self, pos):
        path_length = 0
        for ii in range(len(pos) - 1):
            path_length += np.sqrt(np.sum(abs(pos[ii + 1] - pos[ii]) ** 2))
        return path_length



import signal
import threading
import time

from bec_lib.bec_errors import ScanInterruption

PAUSE_MSG = """
The Scan Queue is entering a paused state. These are your options for changing
the state of the queue:

%resume              Resume the scan.
%restart             Restart the scan.
%abort               Perform cleanup, then kill plan. Mark exit_stats='aborted'.
%halt                Emergency Stop: Do not perform cleanup --- just stop.
"""


class SignalHandler:
    """Context manager for signal handing

    If multiple signals come in quickly, they may not all be seen, quoting
    the libc manual:

      Remember that if there is a particular signal pending for your
      process, additional signals of that same type that arrive in the
      meantime might be discarded. For example, if a SIGINT signal is
      pending when another SIGINT signal arrives, your program will
      probably only see one of them when you unblock this signal.

    https://www.gnu.org/software/libc/manual/html_node/Checking-for-Pending-Signals.html
    """

    def __init__(self, sig, log=None):
        self.sig = sig
        self.interrupted = False
        self.count = 0
        self.log = log
        self.released = False
        self.original_handler = None

    def __enter__(self):
        self.interrupted = False
        self.count = 0

        self.original_handler = signal.getsignal(self.sig)

        def handler(signum, frame):
            self.interrupted = True
            self.count += 1
            if self.log is not None:
                self.log.debug("SignalHandler caught SIGINT; count is %r", self.count)
            if self.count > 10:
                orig_func = self.original_handler
                self.release()
                orig_func(signum, frame)

            self.handle_signals()

        signal.signal(self.sig, handler)
        return self

    def __exit__(self, _type, _value, _tb):
        self.release()

    def release(self):
        if self.released:
            return False
        signal.signal(self.sig, self.original_handler)
        self.released = True
        return True

    def handle_signals(self): ...


class SigintHandler(SignalHandler):
    def __init__(self, bec):
        super().__init__(signal.SIGINT)
        self.bec = bec
        self.last_sigint_time = None  # time most recent SIGINT was processed
        self.num_sigints_processed = 0  # count SIGINTs processed

    def handle_signals(self):
        current_scan = self.bec.queue.scan_storage.current_scan_info
        if not current_scan:
            raise KeyboardInterrupt

        status = current_scan.get("status").lower()
        if status not in ["running", "deferred_pause"]:
            raise KeyboardInterrupt

        if any(current_scan.get("is_scan")) and (
            self.last_sigint_time is None or time.time() - self.last_sigint_time > 10
        ):
            # reset the counter to 1
            # It's been 10 seconds since the last SIGINT. Reset.
            self.count = 1
            if self.last_sigint_time is not None:
                print("It has been 10 seconds since the last SIGINT. Resetting SIGINT handler.")

            threading.Thread(
                target=self.bec.queue.request_scan_interruption, args=(True,), daemon=True
            ).start()
            print(
                "A 'deferred pause' has been requested. The "
                "scan will pause at the next checkpoint. "
                "To pause immediately, hit Ctrl+C again in the "
                "next 10 seconds."
            )

            self.last_sigint_time = time.time()
            return

        # - Ctrl-C twice within 10 seconds or a direct command (e.g. mv) -> hard pause
        if self.bec._service_config.abort_on_ctrl_c:
            print("The scan will be aborted.")
            threading.Thread(target=self.bec.queue.request_scan_abortion, daemon=True).start()
            raise ScanInterruption("User abort.")
        print("A hard pause will be requested.")
        threading.Thread(
            target=self.bec.queue.request_scan_interruption, args=(False,), daemon=True
        ).start()
        raise ScanInterruption(PAUSE_MSG)



# pylint: skip-file
from unittest import mock

import pytest

from bec_lib.bl_checks import (
    BeamlineCheckError,
    BeamlineChecks,
    _run_on_failure,
    _run_with_bl_checks,
)


def test_run_with_bl_checks():
    bl_checks = mock.MagicMock()
    bl_checks._levels = []
    bl_checks._is_paused = False
    _run_with_bl_checks(bl_checks, mock.MagicMock(), 1, 2, 3, a=4, b=5)
    assert not bl_checks._levels


def test_bl_check_raises_on_failed_nested_calls():
    bl_checks = mock.MagicMock()
    bl_checks._levels = [{"fcn": mock.MagicMock()}]
    bl_checks._is_paused = False
    bl_checks.beam_is_okay = False
    with pytest.raises(BeamlineCheckError):
        _run_with_bl_checks(bl_checks, mock.MagicMock(), 1, 2, 3, a=4, b=5)


def test_bl_check_run_on_failure():
    bl_checks = mock.MagicMock()
    bl_checks._levels = []
    bl_checks._is_paused = False
    bl_checks.beam_is_okay = False
    fcn = mock.MagicMock()
    fcn.side_effect = BeamlineCheckError
    _run_on_failure(bl_checks, fcn, 1, 2, 3, a=4, b=5)
    bl_checks.wait_for_beamline_checks.assert_called_once()


def test_bl_check_register():
    client = mock.MagicMock()
    bl_check = BeamlineChecks(client=client)
    condition = mock.MagicMock()
    condition.name = "test"
    bl_check.register(condition)
    assert bl_check.test == condition  # pylint: disable=no-member
    assert bl_check._beamline_checks["test"] == condition


def test_bl_check_pause():
    client = mock.MagicMock()
    bl_check = BeamlineChecks(client=client)
    bl_check.pause()
    assert bl_check._is_paused


def test_bl_check_resume():
    client = mock.MagicMock()
    bl_check = BeamlineChecks(client=client)
    bl_check._is_paused = True
    bl_check.resume()
    assert not bl_check._is_paused


def test_bl_check_reset():
    client = mock.MagicMock()
    bl_check = BeamlineChecks(client=client)
    bl_check._beam_is_okay = False
    bl_check._check_msgs = ["test"]
    bl_check.reset()
    assert bl_check._beam_is_okay
    assert not bl_check._check_msgs


def test_bl_check_disable_check():
    client = mock.MagicMock()
    bl_check = BeamlineChecks(client=client)
    bl_check._beamline_checks = {"test": mock.MagicMock()}
    bl_check.disable_check("test")
    assert not bl_check._beamline_checks["test"].enabled


def test_bl_check_enable_check():
    client = mock.MagicMock()
    bl_check = BeamlineChecks(client=client)
    bl_check._beamline_checks = {"test": mock.MagicMock()}
    bl_check.enable_check("test")
    assert bl_check._beamline_checks["test"].enabled


def test_bl_check_disable_all_checks():
    client = mock.MagicMock()
    bl_check = BeamlineChecks(client=client)
    bl_check._beamline_checks = {"test": mock.MagicMock()}
    bl_check.disable_all_checks()
    assert not bl_check._beamline_checks["test"].enabled


def test_bl_check_enable_all_checks():
    client = mock.MagicMock()
    bl_check = BeamlineChecks(client=client)
    bl_check._beamline_checks = {"test": mock.MagicMock()}
    bl_check.enable_all_checks()
    assert bl_check._beamline_checks["test"].enabled


def test_bl_check_run_beamline_checks():
    client = mock.MagicMock()
    bl_check = BeamlineChecks(client=client)
    bl_check._beamline_checks = {"test": mock.MagicMock()}
    bl_check._beam_is_okay = True
    bl_check._run_beamline_checks()
    assert bl_check._beam_is_okay
    bl_check._beamline_checks["test"].run.assert_called_once()
    bl_check._beamline_checks["test"].on_failure_msg.assert_not_called()


def test_bl_check_send_to_scilog():
    client = mock.MagicMock()
    bl_check = BeamlineChecks(client=client)
    bl_check._send_to_scilog("test")
    client.logbook.send_logbook_message.assert_called_once()


def test_bl_check_beam_status_is_okay():
    client = mock.MagicMock()
    bl_check = BeamlineChecks(client=client)
    with mock.patch.object(bl_check, "_run_beamline_checks") as mock_run:
        mock_run.return_value = []
        assert bl_check._beam_status_is_okay() is True


def test_bl_check_wait_for_beamline_checks():
    client = mock.MagicMock()
    bl_check = BeamlineChecks(client=client)
    bl_check._check_msgs = ["test"]
    bl_check._beam_is_okay = False
    with mock.patch.object(bl_check, "_print_beamline_checks") as mock_print:
        with mock.patch.object(bl_check, "_send_to_scilog") as mock_send:
            with mock.patch.object(bl_check, "_run_beamline_checks") as mock_run:
                mock_run.return_value = ["test"]
                bl_check.wait_for_beamline_checks()
                mock_print.assert_called_once()
                mock_send.assert_called()


def test_bl_check_start():
    client = mock.MagicMock()
    bl_check = BeamlineChecks(client=client)
    bl_check._started = False
    bl_check._beam_is_okay = True
    with mock.patch("bec_lib.bl_checks.threading.Thread") as mock_thread:
        bl_check.start()
        mock_thread.assert_called_once()
        mock_thread.return_value.start.assert_called_once()


def test_bl_check_stop():
    client = mock.MagicMock()
    bl_check = BeamlineChecks(client=client)
    bl_check._started = True
    bl_check._beam_is_okay = True
    bl_check._beam_check_thread = mock.MagicMock()
    bl_check.stop()
    bl_check._beam_check_thread.join.assert_called_once()



from __future__ import annotations

from typing import TYPE_CHECKING

from bec_lib import MessageEndpoints, bec_logger, messages

from .emitter import EmitterBase

logger = bec_logger.logger

if TYPE_CHECKING:
    from .scan_bundler import ScanBundler


class BECEmitter(EmitterBase):
    def __init__(self, scan_bundler: ScanBundler) -> None:
        super().__init__(scan_bundler.connector)
        self.scan_bundler = scan_bundler

    def on_scan_point_emit(self, scan_id: str, point_id: int):
        self._send_bec_scan_point(scan_id, point_id)

    def on_baseline_emit(self, scan_id: str):
        self._send_baseline(scan_id)

    def _send_bec_scan_point(self, scan_id: str, point_id: int) -> None:
        sb = self.scan_bundler

        info = sb.sync_storage[scan_id]["info"]
        msg = messages.ScanMessage(
            point_id=point_id,
            scan_id=scan_id,
            data=sb.sync_storage[scan_id][point_id],
            metadata={
                "scan_id": scan_id,
                "scan_type": info.get("scan_type"),
                "scan_report_devices": info.get("scan_report_devices"),
            },
        )
        self.add_message(
            msg,
            MessageEndpoints.scan_segment(),
            MessageEndpoints.public_scan_segment(scan_id=scan_id, point_id=point_id),
        )

    def _send_baseline(self, scan_id: str) -> None:
        sb = self.scan_bundler

        msg = messages.ScanBaselineMessage(
            scan_id=scan_id,
            data=sb.sync_storage[scan_id]["baseline"],
            metadata=sb.sync_storage[scan_id]["info"],
        )
        pipe = sb.connector.pipeline()
        sb.connector.set(
            MessageEndpoints.public_scan_baseline(scan_id=scan_id), msg, expire=1800, pipe=pipe
        )
        sb.connector.set_and_publish(MessageEndpoints.scan_baseline(), msg, pipe=pipe)
        pipe.execute()



from unittest import mock

import pytest

from bec_lib import messages
from bec_lib.redis_connector import MessageObject
from bec_server.data_processing.dap_service import DAPServiceBase
from bec_server.data_processing.dap_service_manager import DAPServiceManager


class ServiceMock(DAPServiceBase):
    def configure(self, *args, **kwargs):
        pass

    def process(self, *args, **kwargs):
        pass

    def on_scan_status_update(self, *args, **kwargs):
        pass

    @classmethod
    def get_class_doc_string(cls, *args, **kwargs):
        return "class_doc"


@pytest.fixture
def service_manager():
    dap_service_manager = DAPServiceManager(services=ServiceMock)
    dap_service_manager.start(client=mock.MagicMock())
    yield dap_service_manager
    dap_service_manager.shutdown()


def test_DAPServiceManager_init(service_manager):
    assert service_manager.services == [ServiceMock]
    assert service_manager.available_dap_services == {
        "ServiceMock": {
            "class": "ServiceMock",
            "user_friendly_name": "ServiceMock",
            "class_doc": "class_doc",
            "run_doc": None,
            "run_name": "run",
            "signature": [
                {
                    "name": "args",
                    "kind": "VAR_POSITIONAL",
                    "default": "_empty",
                    "annotation": "_empty",
                },
                {
                    "name": "kwargs",
                    "kind": "VAR_KEYWORD",
                    "default": "_empty",
                    "annotation": "_empty",
                },
            ],
            "params": {},
            "auto_run_supported": False,
            "class_args": [],
            "class_kwargs": {},
        }
    }


@pytest.mark.parametrize(
    "msg, process_called",
    [
        (messages.DAPRequestMessage(dap_cls="dap_cls", dap_type="on_demand", config={}), True),
        (messages.ScanStatusMessage(scan_id="scan_id", status="open", info={}), False),
    ],
)
def test_DAPServiceManager_request_callback(service_manager, msg, process_called):
    msg_obj = MessageObject(value=msg, topic="topic")
    with mock.patch.object(service_manager, "process_dap_request") as mock_process_dap_request:
        service_manager._dap_request_callback(msg_obj)
        if process_called:
            mock_process_dap_request.assert_called_once_with(msg)


@pytest.mark.parametrize(
    "msg, raised_exception, error_msg",
    [
        (
            messages.DAPRequestMessage(
                dap_cls="ServiceMock",
                dap_type="continuous",
                config={"auto_fit": True, "class_args": [], "class_kwargs": {}},
            ),
            False,
            "",
        ),
        (
            messages.DAPRequestMessage(
                dap_cls="ServiceMock", dap_type="continuous", config={"auto_fit": False}
            ),
            False,
            "",
        ),
        (
            messages.DAPRequestMessage(dap_cls="ServiceMock", dap_type="on_demand", config={}),
            False,
            "",
        ),
    ],
)
def test_process_dap_request(service_manager, msg, raised_exception, error_msg):
    if raised_exception:
        with mock.patch.object(service_manager, "send_dap_response") as mock_send_dap_response:
            service_manager.process_dap_request(msg)
            mock_send_dap_response.assert_called_once_with(
                msg, success=False, error=error_msg, metadata=msg.metadata
            )
        return

    service_manager.process_dap_request(msg)



from __future__ import annotations

import builtins
import functools
import os
import time
import uuid
from typing import TYPE_CHECKING

import yaml

import bec_lib
from bec_lib import BECClient, messages
from bec_lib.connector import ConnectorBase
from bec_lib.devicemanager import DeviceManagerBase
from bec_lib.endpoints import EndpointInfo, MessageEndpoints
from bec_lib.logger import bec_logger
from bec_lib.scans import Scans

if TYPE_CHECKING:
    from bec_lib.alarm_handler import Alarms

dir_path = os.path.dirname(bec_lib.__file__)

logger = bec_logger.logger

# pylint: disable=no-member
# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name
# pylint: disable=protected-access


def queue_is_empty(queue) -> bool:  # pragma: no cover
    if not queue:
        return True
    if not queue["primary"].get("info"):
        return True
    return False


def get_queue(bec):  # pragma: no cover
    return bec.queue.connector.get(MessageEndpoints.scan_queue_status())


def wait_for_empty_queue(bec):  # pragma: no cover
    while not get_queue(bec):
        time.sleep(1)
    while not queue_is_empty(get_queue(bec).content["queue"]):
        time.sleep(1)
        logger.info(bec.queue)
    while get_queue(bec).content["queue"]["primary"]["status"] != "RUNNING":
        time.sleep(1)
        logger.info(bec.queue)


class ScansMock(Scans):
    def _import_scans(self):
        pass

    def open_scan_def(self):
        pass

    def close_scan_def(self):
        pass

    def close_scan_group(self):
        pass

    def umv(self):
        pass


class ClientMock(BECClient):
    def _load_scans(self):
        self.scans = ScansMock(self)
        builtins.scans = self.scans

    # def _start_services(self):
    #     pass

    def _start_metrics_emitter(self):
        pass

    def _start_update_service_info(self):
        pass


def get_device_info_mock(device_name, device_class) -> messages.DeviceInfoMessage:
    device_info = {
        "samx": messages.DeviceInfoMessage(
            device="samx",
            info={
                "device_info": {
                    "device_base_class": "positioner",
                    "signals": [
                        {
                            "component_name": "readback",
                            "obj_name": device_name,
                            "kind_int": 5,
                            "kind_str": "Kind.hinted",
                        },
                        {
                            "component_name": "setpoint",
                            "obj_name": f"{device_name}_setpoint",
                            "kind_int": 1,
                            "kind_str": "Kind.normal",
                        },
                        {
                            "component_name": "motor_is_moving",
                            "obj_name": f"{device_name}_motor_is_moving",
                            "kind_int": 1,
                            "kind_str": "Kind.normal",
                        },
                        {
                            "component_name": "readback",
                            "obj_name": device_name,
                            "kind_int": 5,
                            "kind_str": "Kind.hinted",
                        },
                        {
                            "component_name": "velocity",
                            "obj_name": f"{device_name}_velocity",
                            "kind_int": 2,
                            "kind_str": "Kind.config",
                        },
                        {
                            "component_name": "acceleration",
                            "obj_name": f"{device_name}_acceleration",
                            "kind_int": 2,
                            "kind_str": "Kind.config",
                        },
                        {
                            "component_name": "high_limit_travel",
                            "obj_name": f"{device_name}_high_limit_travel",
                            "kind_int": 2,
                            "kind_str": "Kind.config",
                        },
                        {
                            "component_name": "low_limit_travel",
                            "obj_name": f"{device_name}_low_limit_travel",
                            "kind_int": 2,
                            "kind_str": "Kind.config",
                        },
                        {
                            "component_name": "unused",
                            "obj_name": f"{device_name}_unused",
                            "kind_int": 0,
                            "kind_str": "Kind.omitted",
                        },
                    ],
                    "hints": {"fields": ["samx"]},
                    "describe": {
                        "samx": {
                            "source": "SIM:samx",
                            "dtype": "integer",
                            "shape": [],
                            "precision": 3,
                        },
                        "samx_setpoint": {
                            "source": "SIM:samx_setpoint",
                            "dtype": "integer",
                            "shape": [],
                            "precision": 3,
                        },
                        "samx_motor_is_moving": {
                            "source": "SIM:samx_motor_is_moving",
                            "dtype": "integer",
                            "shape": [],
                            "precision": 3,
                        },
                    },
                    "describe_configuration": {
                        "samx_velocity": {
                            "source": "SIM:samx_velocity",
                            "dtype": "integer",
                            "shape": [],
                        },
                        "samx_acceleration": {
                            "source": "SIM:samx_acceleration",
                            "dtype": "integer",
                            "shape": [],
                        },
                    },
                    "sub_devices": [],
                    "custom_user_access": {
                        "dummy_controller": {
                            "_func_with_args": {"type": "func", "doc": None},
                            "_func_with_args_and_kwargs": {"type": "func", "doc": None},
                            "_func_with_kwargs": {"type": "func", "doc": None},
                            "_func_without_args_kwargs": {"type": "func", "doc": None},
                            "controller_show_all": {
                                "type": "func",
                                "doc": (
                                    "dummy controller show all\n\n        Raises:\n           "
                                    " in: _description_\n            LimitError:"
                                    " _description_\n\n        Returns:\n            _type_:"
                                    " _description_\n        "
                                ),
                            },
                            "some_var": {"type": "int"},
                        },
                        "sim_state": {"type": "dict"},
                        "speed": {"type": "int"},
                    },
                }
            },
        ),
        "dyn_signals": messages.DeviceInfoMessage(
            device="dyn_signals",
            info={
                "device_info": {
                    "device_dotted_name": "dyn_signals",
                    "device_attr_name": "dyn_signals",
                    "device_base_class": "device",
                    "signals": [],
                    "hints": {"fields": []},
                    "describe": {
                        "dyn_signals_messages_message1": {
                            "source": "SIM:dyn_signals_messages_message1",
                            "dtype": "integer",
                            "shape": [],
                            "precision": 3,
                        },
                        "dyn_signals_messages_message2": {
                            "source": "SIM:dyn_signals_messages_message2",
                            "dtype": "integer",
                            "shape": [],
                            "precision": 3,
                        },
                        "dyn_signals_messages_message3": {
                            "source": "SIM:dyn_signals_messages_message3",
                            "dtype": "integer",
                            "shape": [],
                            "precision": 3,
                        },
                        "dyn_signals_messages_message4": {
                            "source": "SIM:dyn_signals_messages_message4",
                            "dtype": "integer",
                            "shape": [],
                            "precision": 3,
                        },
                        "dyn_signals_messages_message5": {
                            "source": "SIM:dyn_signals_messages_message5",
                            "dtype": "integer",
                            "shape": [],
                            "precision": 3,
                        },
                    },
                    "describe_configuration": {},
                    "sub_devices": [
                        {
                            "device_name": "dyn_signals_messages",
                            "device_info": {
                                "device_attr_name": "messages",
                                "device_dotted_name": "messages",
                                "device_base_class": "device",
                                "signals": [
                                    {
                                        "component_name": "message1",
                                        "obj_name": "dyn_signals_messages_message1",
                                        "kind_int": 1,
                                        "kind_str": "Kind.normal",
                                    },
                                    {
                                        "component_name": "message2",
                                        "obj_name": "dyn_signals_messages_message2",
                                        "kind_int": 1,
                                        "kind_str": "Kind.normal",
                                    },
                                    {
                                        "component_name": "message3",
                                        "obj_name": "dyn_signals_messages_message3",
                                        "kind_int": 1,
                                        "kind_str": "Kind.normal",
                                    },
                                    {
                                        "component_name": "message4",
                                        "obj_name": "dyn_signals_messages_message4",
                                        "kind_int": 1,
                                        "kind_str": "Kind.normal",
                                    },
                                    {
                                        "component_name": "message5",
                                        "obj_name": "dyn_signals_messages_message5",
                                        "kind_int": 1,
                                        "kind_str": "Kind.normal",
                                    },
                                ],
                                "hints": {"fields": []},
                                "describe": {
                                    "dyn_signals_messages_message1": {
                                        "source": "SIM:dyn_signals_messages_message1",
                                        "dtype": "integer",
                                        "shape": [],
                                        "precision": 3,
                                    },
                                    "dyn_signals_messages_message2": {
                                        "source": "SIM:dyn_signals_messages_message2",
                                        "dtype": "integer",
                                        "shape": [],
                                        "precision": 3,
                                    },
                                    "dyn_signals_messages_message3": {
                                        "source": "SIM:dyn_signals_messages_message3",
                                        "dtype": "integer",
                                        "shape": [],
                                        "precision": 3,
                                    },
                                    "dyn_signals_messages_message4": {
                                        "source": "SIM:dyn_signals_messages_message4",
                                        "dtype": "integer",
                                        "shape": [],
                                        "precision": 3,
                                    },
                                    "dyn_signals_messages_message5": {
                                        "source": "SIM:dyn_signals_messages_message5",
                                        "dtype": "integer",
                                        "shape": [],
                                        "precision": 3,
                                    },
                                },
                                "describe_configuration": {},
                                "sub_devices": [],
                                "custom_user_access": {},
                            },
                        }
                    ],
                    "custom_user_access": {},
                }
            },
        ),
    }
    if device_name in device_info:
        return device_info[device_name]

    device_base_class = "positioner" if device_class == "SimPositioner" else "signal"
    if device_base_class == "positioner":
        signals = [
            {
                "component_name": "readback",
                "obj_name": device_name,
                "kind_int": 5,
                "kind_str": "Kind.hinted",
            },
            {
                "component_name": "setpoint",
                "obj_name": f"{device_name}_setpoint",
                "kind_int": 1,
                "kind_str": "Kind.normal",
            },
            {
                "component_name": "motor_is_moving",
                "obj_name": f"{device_name}_motor_is_moving",
                "kind_int": 1,
                "kind_str": "Kind.normal",
            },
            {
                "component_name": "readback",
                "obj_name": device_name,
                "kind_int": 5,
                "kind_str": "Kind.hinted",
            },
            {
                "component_name": "velocity",
                "obj_name": f"{device_name}_velocity",
                "kind_int": 2,
                "kind_str": "Kind.config",
            },
            {
                "component_name": "acceleration",
                "obj_name": f"{device_name}_acceleration",
                "kind_int": 2,
                "kind_str": "Kind.config",
            },
            {
                "component_name": "high_limit_travel",
                "obj_name": f"{device_name}_high_limit_travel",
                "kind_int": 2,
                "kind_str": "Kind.config",
            },
            {
                "component_name": "low_limit_travel",
                "obj_name": f"{device_name}_low_limit_travel",
                "kind_int": 2,
                "kind_str": "Kind.config",
            },
            {
                "component_name": "unused",
                "obj_name": f"{device_name}_unused",
                "kind_int": 0,
                "kind_str": "Kind.omitted",
            },
        ]
    elif device_base_class == "signal":
        signals = [
            {
                "component_name": "readback",
                "obj_name": device_name,
                "kind_int": 5,
                "kind_str": "Kind.hinted",
            },
            {
                "component_name": "velocity",
                "obj_name": f"{device_name}_velocity",
                "kind_int": 2,
                "kind_str": "Kind.config",
            },
            {
                "component_name": "acceleration",
                "obj_name": f"{device_name}_acceleration",
                "kind_int": 2,
                "kind_str": "Kind.config",
            },
            {
                "component_name": "high_limit_travel",
                "obj_name": f"{device_name}_high_limit_travel",
                "kind_int": 2,
                "kind_str": "Kind.config",
            },
            {
                "component_name": "low_limit_travel",
                "obj_name": f"{device_name}_low_limit_travel",
                "kind_int": 2,
                "kind_str": "Kind.config",
            },
            {
                "component_name": "unused",
                "obj_name": f"{device_name}_unused",
                "kind_int": 0,
                "kind_str": "Kind.omitted",
            },
        ]
    dev_info = {
        "device_name": device_name,
        "device_info": {
            "device_dotted_name": device_name,
            "device_attr_name": device_name,
            "device_base_class": device_base_class,
            "signals": signals,
        },
        "custom_user_acces": {},
    }

    return messages.DeviceInfoMessage(device=device_name, info=dev_info, metadata={})


class DMClientMock(DeviceManagerBase):
    def _get_device_info(self, device_name) -> messages.DeviceInfoMessage:
        return get_device_info_mock(device_name, self.get_device(device_name)["deviceClass"])

    def get_device(self, device_name):
        for dev in self._session["devices"]:
            if dev["name"] == device_name:
                return dev


class PipelineMock:  # pragma: no cover
    _pipe_buffer = []
    _connector = None

    def __init__(self, connector) -> None:
        self._connector = connector

    def execute(self):
        if not self._connector.store_data:
            self._pipe_buffer = []
            return []
        res = [
            getattr(self._connector, method)(*args, **kwargs)
            for method, args, kwargs in self._pipe_buffer
        ]
        self._pipe_buffer = []
        return res


class SignalMock:  # pragma: no cover
    def __init__(self) -> None:
        self.is_set = False

    def set(self):
        self.is_set = True


class ConnectorMock(ConnectorBase):  # pragma: no cover
    def __init__(self, bootstrap_server="localhost:0000", store_data=True):
        super().__init__(bootstrap_server)
        self.message_sent = []
        self._get_buffer = {}
        self.store_data = store_data

    def raise_alarm(
        self, severity: Alarms, alarm_type: str, source: str, msg: dict, metadata: dict
    ):
        pass

    def log_error(self, *args, **kwargs):
        pass

    def shutdown(self):
        pass

    def register(self, *args, **kwargs):
        pass

    def unregister(self, *args, **kwargs):
        pass

    def poll_messages(self, *args, **kwargs):
        pass

    def keys(self, *args, **kwargs):
        return []

    def set(self, topic, msg, pipe=None, expire: int = None):
        if pipe:
            pipe._pipe_buffer.append(("set", (topic.endpoint, msg), {"expire": expire}))
            return
        self.message_sent.append({"queue": topic, "msg": msg, "expire": expire})

    def raw_send(self, topic, msg, pipe=None):
        if pipe:
            pipe._pipe_buffer.append(("send", (topic.endpoint, msg), {}))
            return
        self.message_sent.append({"queue": topic, "msg": msg})

    def send(self, topic, msg, pipe=None):
        if not isinstance(msg, messages.BECMessage):
            raise TypeError("Message must be a BECMessage")
        return self.raw_send(topic, msg, pipe)

    def set_and_publish(self, topic, msg, pipe=None, expire: int = None):
        if pipe:
            pipe._pipe_buffer.append(("set_and_publish", (topic.endpoint, msg), {"expire": expire}))
            return
        self.message_sent.append({"queue": topic, "msg": msg, "expire": expire})

    def lpush(self, topic, msg, pipe=None, max_size=None):
        if pipe:
            pipe._pipe_buffer.append(("lpush", (topic, msg), {}))
            return

    def rpush(self, topic, msg, pipe=None):
        if pipe:
            pipe._pipe_buffer.append(("rpush", (topic, msg), {}))
            return
        pass

    def lrange(self, topic, start, stop, pipe=None):
        if pipe:
            pipe._pipe_buffer.append(("lrange", (topic, start, stop), {}))
            return
        return []

    def get(self, topic, pipe=None):
        if isinstance(topic, EndpointInfo):
            topic = topic.endpoint
        if pipe:
            pipe._pipe_buffer.append(("get", (topic,), {}))
            return
        val = self._get_buffer.get(topic)
        if isinstance(val, list):
            return val.pop(0)
        self._get_buffer.pop(topic, None)
        return val

    def pipeline(self):
        return PipelineMock(self)

    def delete(self, topic, pipe=None):
        if pipe:
            pipe._pipe_buffer.append(("delete", (topic,), {}))
            return

    def lset(self, topic: str, index: int, msgs: str, pipe=None) -> None:
        if pipe:
            pipe._pipe_buffer.append(("lrange", (topic, index, msgs), {}))
            return

    def producer(self):
        return self

    def execute_pipeline(self, pipeline):
        pipeline.execute()

    def xadd(self, topic, msg_dict, max_size=None, pipe=None, expire: int = None):
        if pipe:
            pipe._pipe_buffer.append(("xadd", (topic, msg_dict), {"expire": expire}))
            return
        pass

    def xread(self, topic, id=None, count=None, block=None, pipe=None, from_start=False):
        if pipe:
            pipe._pipe_buffer.append(
                ("xread", (topic, id, count, block), {"from_start": from_start})
            )
            return
        return []

    def xrange(self, topic, min="-", max="+", pipe=None):
        if pipe:
            pipe._pipe_buffer.append(("xrange", (topic, min, max), {}))
            return
        return []



"""
This module contains the DeviceManager class which is used to manage devices and their configuration.
"""

from __future__ import annotations

import re
import traceback
from typing import TYPE_CHECKING

from rich.console import Console
from rich.table import Table
from typeguard import typechecked

from bec_lib import messages
from bec_lib.bec_errors import DeviceConfigError
from bec_lib.config_helper import ConfigHelper
from bec_lib.device import ComputedSignal, Device, DeviceBase, Positioner, ReadoutPriority, Signal
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from bec_lib.messages import BECStatus, DeviceConfigMessage, DeviceInfoMessage

if TYPE_CHECKING:
    from bec_lib import BECService

logger = bec_logger.logger


class DeviceContainer(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    self[k] = v

        if kwargs:
            for k, v in kwargs.items():
                self[k] = v

    def __getattr__(self, attr):
        if attr.startswith("_"):
            # if dunder attributes are would not be caught, they
            # would raise a DeviceConfigError and kill the
            # IPython completer
            return self.get(attr)
        dev = self.get(attr)
        if not dev:
            raise DeviceConfigError(f"Device {attr} does not exist.")
        return dev

    def __setattr__(self, key, value):
        if isinstance(value, DeviceBase):
            self.__setitem__(key, value)
        else:
            raise AttributeError("Unsupported device type.")

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super().__delitem__(key)
        del self.__dict__[key]

    def flush(self) -> None:
        """
        Remove all devices from the device manager
        """
        self.clear()
        self.__dict__.clear()

    @property
    def enabled_devices(self) -> list:
        """get a list of enabled devices"""
        return [dev for _, dev in self.items() if dev.enabled]

    @property
    def disabled_devices(self) -> list:
        """get a list of disabled devices"""
        return [dev for _, dev in self.items() if not dev.enabled]

    def readout_priority(self, readout_priority: ReadoutPriority) -> list:
        """get all devices with the specified readout proprity

        Args:
            readout_priority (str): Readout priority (e.g.  on_request, baseline, monitored, async, continuous)

        Returns:
            list: List of devices that belong to the specified acquisition readoutPriority
        """
        val = ReadoutPriority(readout_priority)
        # pylint: disable=protected-access
        return [dev for _, dev in self.items() if dev._config["readoutPriority"] == val]

    def _filter_devices(
        self, readout_priority: ReadoutPriority, readout_priority_mod: dict, devices: list = None
    ) -> list:
        """filter devices by readout priority"""
        if devices is None:
            devices = []

        if readout_priority_mod is None:
            readout_priority_mod = {}

        devices = [self.get(dev) if isinstance(dev, str) else dev for dev in devices]

        devices.extend(self.readout_priority(readout_priority))
        devices.extend(
            [self.get(dev) for dev in readout_priority_mod.get(readout_priority.name.lower(), [])]
        )

        excluded_readout_priority = [
            str(x.name).lower() for x in ReadoutPriority if x != readout_priority
        ]
        excluded_devices = self.disabled_devices
        for priority in excluded_readout_priority:
            excluded_devices.extend(self.get(dev) for dev in readout_priority_mod.get(priority, []))

        return [dev for dev in set(devices) if dev not in excluded_devices]

    def async_devices(self, readout_priority: dict | None = None) -> list:
        """get a list of all synchronous devices"""
        # pylint: disable=protected-access
        return self._filter_devices(ReadoutPriority.ASYNC, readout_priority)

    def continuous_devices(self, readout_priority: dict | None = None) -> list:
        """get a list of all continuous devices"""
        # pylint: disable=protected-access
        return self._filter_devices(ReadoutPriority.CONTINUOUS, readout_priority)

    def on_request_devices(self, readout_priority: dict | None = None) -> list:
        """get a list of all on request devices"""
        # pylint: disable=protected-access
        return self._filter_devices(ReadoutPriority.ON_REQUEST, readout_priority)

    @typechecked
    def monitored_devices(
        self, scan_motors: list | None = None, readout_priority: dict | None = None
    ) -> list:
        """get a list of all enabled primary devices"""
        devices = []
        if scan_motors:
            if not isinstance(scan_motors, list):
                scan_motors = [scan_motors]
            for scan_motor in scan_motors:
                if scan_motor not in devices:
                    if isinstance(scan_motor, DeviceBase):
                        devices.append(scan_motor)
                    else:
                        devices.append(self.get(scan_motor))

        return self._filter_devices(ReadoutPriority.MONITORED, readout_priority, devices)

    @typechecked
    def baseline_devices(
        self, scan_motors: list | None = None, readout_priority: dict | None = None
    ) -> list:
        """
        Get a list of all enabled baseline devices
        Args:
            scan_motors(list): list of scan motors
            readout_priority(dict): readout priority

        Returns:
            list: list of baseline devices
        """
        devices = self.readout_priority(ReadoutPriority.BASELINE)
        if scan_motors:
            if not isinstance(scan_motors, list):
                scan_motors = [scan_motors]
            for scan_motor in scan_motors:
                if scan_motor not in devices:
                    if isinstance(scan_motor, DeviceBase):
                        devices.append(scan_motor)
                    else:
                        devices.append(self.get(scan_motor))
            excluded_devices = scan_motors
        else:
            excluded_devices = []
        if readout_priority is None:
            readout_priority = {}

        devices.extend([self.get(dev) for dev in readout_priority.get("baseline", [])])
        excluded_devices.extend(self.disabled_devices)
        excluded_devices.extend([self.get(dev) for dev in readout_priority.get("monitored", [])])
        excluded_devices.extend([self.get(dev) for dev in readout_priority.get("on_request", [])])
        excluded_devices.extend([self.get(dev) for dev in readout_priority.get("continuous", [])])
        excluded_devices.extend([self.get(dev) for dev in readout_priority.get("async", [])])

        return [dev for dev in set(devices) if dev not in excluded_devices]

    def get_devices_with_tags(self, tags: list) -> list:
        """
        Get a list of all devices with the specified tags
        Args:
            tags (list): List of tags

        Returns:
            list: List of devices with the specified tags
        """
        # pylint: disable=protected-access
        if not isinstance(tags, list):
            tags = [tags]
        return [
            dev for _, dev in self.items() if set(tags) & set(dev._config.get("deviceTags", []))
        ]

    def show_tags(self) -> list:
        """returns a list of used tags in the current config"""
        tags = set()
        for _, dev in self.items():
            # pylint: disable=protected-access
            dev_tags = dev._config.get("deviceTags")
            if dev_tags:
                tags.update(dev_tags)
        return list(tags)

    def get_software_triggered_devices(self) -> list:
        """get a list of all devices that should receive a software trigger detectors"""
        # pylint: disable=protected-access
        devices = [
            dev for _, dev in self.items() if dev._config.get("softwareTrigger", False) is True
        ]
        excluded_devices = self.disabled_devices
        return [dev for dev in set(devices) if dev not in excluded_devices]

    def _expand_device_name(self, device_name: str) -> list[str]:
        try:
            regex = re.compile(device_name)
        except re.error:
            return [device_name]
        return [dev for dev in self.keys() if regex.match(dev)]

    def wm(self, device_names: list[str | DeviceBase | None] = None, *args):
        """Get the current position of one or more devices.

        Args:
            device_names (list): List of device names or Device objects

        Examples:
            >>> dev.wm()
            >>> dev.wm('sam*')
            >>> dev.wm('samx')
            >>> dev.wm(['samx', 'samy'])
            >>> dev.wm(dev.monitored_devices())
            >>> dev.wm(dev.get_devices_with_tags('user motors'))

        """
        if not device_names:
            device_names = self.values()
        else:
            expanded_devices = []
            if not isinstance(device_names, list):
                device_names = [device_names]
            if len(device_names) == 0:
                return

            for dev in device_names:
                if isinstance(dev, DeviceBase):
                    expanded_devices.append(dev)
                else:
                    devs = self._expand_device_name(dev)
                    expanded_devices.extend([self.__dict__[dev] for dev in devs])
            device_names = expanded_devices
        console = Console()
        table = Table()
        table.add_column("", justify="center")
        table.add_column("readback", justify="center")
        table.add_column("setpoint", justify="center")
        table.add_column("limits", justify="center")
        dev_read = {dev.name: dev.read(cached=True) for dev in device_names}
        readbacks = {}
        setpoints = {}
        limits = {}
        for dev in device_names:
            if hasattr(dev, "limits"):
                limits[dev.name] = str(dev.limits)
            else:
                limits[dev.name] = "[]"

        for dev, read in dev_read.items():
            if dev in read:
                val = read[dev]["value"]
                if not isinstance(val, str):
                    readbacks[dev] = f"{val:.4f}"
                else:
                    readbacks[dev] = val
            else:
                readbacks[dev] = "N/A"
            if f"{dev}_setpoint" in read:
                val = read[f"{dev}_setpoint"]["value"]
                if not isinstance(val, str):
                    setpoints[dev] = f"{val:.4f}"
                else:
                    setpoints[dev] = val
            elif f"{dev}_user_setpoint" in read:
                val = read[f"{dev}_user_setpoint"]["value"]
                if not isinstance(val, str):
                    setpoints[dev] = f"{val:.4f}"
                else:
                    setpoints[dev] = val
            else:
                setpoints[dev] = "N/A"
        for dev in device_names:
            table.add_row(dev.name, readbacks[dev.name], setpoints[dev.name], limits[dev.name])
        console.print(table)

    def _add_device(self, name, obj) -> None:
        """
        Add device a new device to the device manager
        Args:
            name: name of the device
            obj: instance of the device

        Returns:

        """
        self[name] = obj

    def describe(self) -> list:
        """
        Describe all devices associated with the DeviceManager
        Returns:

        """
        return [dev.describe() for name, dev in self.devices.items()]

    def show_all(self, console: Console = None) -> None:
        """print all devices"""

        if console is None:
            console = Console()
        table = Table()
        table.add_column("Device", justify="center")
        table.add_column("Description", justify="center")
        table.add_column("Status", justify="center")
        table.add_column("ReadOnly", justify="center")
        table.add_column("SoftwareTrigger", justify="center")
        table.add_column("Device class", justify="center")
        table.add_column("Readout priority", justify="center")
        table.add_column("Device tags", justify="center")

        # pylint: disable=protected-access
        for dev in self.values():
            table.add_row(
                dev.name,
                dev._config.get("description", dev.name),
                "enabled" if dev.enabled else "disabled",
                str(dev.read_only),
                str(dev.software_trigger),
                dev._config.get("deviceClass"),
                dev._config.get("readoutPriority"),
                str(dev._config.get("deviceTags", [])),
            )
        console.print(table)

    def __str__(self) -> str:
        return "Device container."


class DeviceManagerBase:
    devices = DeviceContainer()
    _config = {}  # valid config
    _session = {}
    _request = None  # requested config
    _request_config_parsed = None  # parsed config request
    _response = None  # response message

    _connector_base_register = {}
    config_helper = None
    _device_cls = DeviceBase
    _status_cb = []

    def __init__(self, service: BECService, status_cb: list = None) -> None:
        self._service = service
        self.parent = service  # for backwards compatibility; will be removed in the future
        self.connector = self._service.connector
        self.config_helper = ConfigHelper(self.connector, self._service._service_name)
        self._status_cb = status_cb if isinstance(status_cb, list) else [status_cb]

    @property
    def producer(self):
        # will show the deprecation warning
        return self.connector.producer()

    def initialize(self, bootstrap_server) -> None:
        """
        Initialize the DeviceManager by starting all connectors.
        Args:
            bootstrap_server: Redis server address, e.g. 'localhost:6379'

        Returns:

        """
        self._start_connectors(bootstrap_server)
        try:
            self._get_config()
        except DeviceConfigError as dev_conf_error:
            logger.error(f"Failed to initialize DeviceManager. {dev_conf_error}")

    def update_status(self, status: BECStatus):
        """Update the status of the device manager
        Args:
            status (BECStatus): New status
        """
        for cb in self._status_cb:
            if cb:
                cb(status)

    def parse_config_message(self, msg: DeviceConfigMessage) -> None:
        """
        Parse a config message and update the device config accordingly.

        Args:
            msg (DeviceConfigMessage): Config message

        """
        # pylint: disable=protected-access
        action = msg.content["action"]
        config = msg.content["config"]
        self.update_status(BECStatus.BUSY)
        if action == "update":
            for dev in config:
                if "deviceConfig" in config[dev]:
                    logger.info(f"Updating device config for device {dev}.")
                    self.devices[dev]._config["deviceConfig"].update(config[dev]["deviceConfig"])
                    logger.debug(
                        f"New config for device {dev}: {self.devices[dev]._config['deviceConfig']}"
                    )
                if "enabled" in config[dev]:
                    self.devices[dev]._config["enabled"] = config[dev]["enabled"]
                    status = "enabled" if self.devices[dev].enabled else "disabled"
                    logger.info(f"Device {dev} has been {status}.")
                if "readOnly" in config[dev]:
                    self.devices[dev]._config["readOnly"] = config[dev]["readOnly"]
                if "userParameter" in config[dev]:
                    self.devices[dev]._config["userParameter"] = config[dev]["userParameter"]
                if "onFailure" in config[dev]:
                    self.devices[dev]._config["onFailure"] = config[dev]["onFailure"]
                if "deviceTags" in config[dev]:
                    self.devices[dev]._config["deviceTags"] = config[dev]["deviceTags"]
                if "readoutPriority" in config[dev]:
                    self.devices[dev]._config["readoutPriority"] = config[dev]["readoutPriority"]
                if "softwareTrigger" in config[dev]:
                    self.devices[dev]._config["softwareTrigger"] = config[dev]["softwareTrigger"]

        elif action == "add":
            self._add_action(config)
        elif action == "reload":
            self._reload_action()
        elif action == "remove":
            self._remove_action(config)
        self.update_status(BECStatus.RUNNING)
        self._acknowledge_config_request(msg)

    def _acknowledge_config_request(self, msg: DeviceConfigMessage) -> None:
        """
        Acknowledge a config request by sending a response message.
        Args:
            msg (DeviceConfigMessage): Config message

        Returns:

        """
        if not msg.metadata.get("RID"):
            return
        self.connector.lpush(
            MessageEndpoints.service_response(msg.metadata["RID"]),
            messages.ServiceResponseMessage(
                # pylint: disable=no-member
                response={"accepted": True, "service": self._service._service_name}
            ),
            expire=100,
        )

    def _add_action(self, config) -> None:
        for dev in config:
            self._add_device(dev, config)

    def _reload_action(self) -> None:
        logger.info("Reloading config.")
        self.devices.flush()
        self._get_config()

    def _remove_action(self, config) -> None:
        for dev in config:
            self._remove_device(dev)

    def _start_connectors(self, bootstrap_server) -> None:
        self._start_base_register()

    def _start_base_register(self) -> None:
        """
        Start consuming messages for all base topics. This method will be called upon startup.
        Returns:

        """
        self.connector.register(
            MessageEndpoints.device_config_update(),
            cb=self._device_config_update_callback,
            parent=self,
        )

    @staticmethod
    def _log_callback(msg, *, parent, **kwargs) -> None:
        """
        Consumer callback for handling log messages.
        Args:
            cls: Reference to the DeviceManager instance
            msg: log message of type LogMessage
            **kwargs: Additional keyword arguments for the callback function

        Returns:

        """
        logger.info(f"Received log message: {str(msg)}")

    @staticmethod
    def _device_config_update_callback(msg, *, parent, **kwargs) -> None:
        """
        Consumer callback for handling new device configuration
        Args:
            cls: Reference to the DeviceManager instance
            msg: message of type DeviceConfigMessage

        Returns:

        """
        logger.info(f"Received new config: {str(msg)}")
        parent.parse_config_message(msg.value)

    def _get_config(self):
        self._session["devices"] = self._get_redis_device_config()
        if not self._session["devices"]:
            logger.warning("No config available.")
        self._load_session()

    def _get_redis_device_config(self) -> list:
        devices = self.connector.get(MessageEndpoints.device_config())
        if not devices:
            return []
        return devices.content["resource"]

    def _add_device(self, dev: dict, msg: messages.DeviceInfoMessage):
        name = msg.content["device"]
        info = msg.content["info"]

        base_class = info["device_info"]["device_base_class"]

        if base_class == "device":
            logger.info(f"Adding new device {name}")
            obj = Device(name=name, info=info, parent=self)
        elif base_class == "positioner":
            logger.info(f"Adding new positioner {name}")
            obj = Positioner(name=name, info=info, parent=self)
        elif base_class == "signal":
            logger.info(f"Adding new signal {name}")
            obj = Signal(name=name, info=info, parent=self)
        elif base_class == "computed_signal":
            logger.info(f"Adding new computed signal {name}")
            obj = ComputedSignal(name=name, info=info, parent=self)
        else:
            logger.error(f"Trying to add new device {name} of type {base_class}")

        # pylint: disable=protected-access
        obj._config = dev
        self.devices._add_device(name, obj)

    def _remove_device(self, dev_name):
        if dev_name in self.devices:
            self.devices.pop(dev_name)

    def _load_session(self, _device_cls=None):
        if self._is_config_valid():
            for dev in self._session["devices"]:
                # pylint: disable=broad-except
                try:
                    msg = self._get_device_info(dev.get("name"))
                    self._add_device(dev, msg)
                except Exception:
                    content = traceback.format_exc()
                    logger.error(f"Failed to load device {dev}: {content}")

    def _get_device_info(self, device_name) -> DeviceInfoMessage:
        return self.connector.get(MessageEndpoints.device_info(device_name))

    def check_request_validity(self, msg: DeviceConfigMessage) -> None:
        """
        Check if the config request is valid.

        Args:
            msg (DeviceConfigMessage): Config message

        """
        if msg.content["action"] not in ["update", "add", "remove", "reload", "set"]:
            raise DeviceConfigError("Action must be either add, remove, update, set or reload.")
        if msg.content["action"] in ["update", "add", "remove", "set"]:
            if not msg.content["config"]:
                raise DeviceConfigError(
                    "Config cannot be empty for an action of type add, remove, set or update."
                )
            if not isinstance(msg.content["config"], dict):
                raise DeviceConfigError("Config must be of type dict.")
        if msg.content["action"] in ["update", "remove"]:
            for dev in msg.content["config"].keys():
                if dev not in self.devices:
                    raise DeviceConfigError(
                        f"Device {dev} does not exist and cannot be updated / removed."
                    )
        if msg.content["action"] == "add":
            for dev in msg.content["config"].keys():
                if dev in self.devices:
                    raise DeviceConfigError(f"Device {dev} already exists and cannot be added.")

    def _is_config_valid(self) -> bool:
        if self._config is None:
            return False
        if not isinstance(self._config, dict):
            return False
        return True

    def shutdown(self):
        """
        Shutdown all connectors.
        """
        self.connector.shutdown()

    def __del__(self):
        self.shutdown()



import os
from unittest import mock

import pytest
import yaml

import bec_lib
from bec_lib import DeviceManagerBase, ServiceConfig, bec_logger, messages
from bec_lib.messages import BECStatus
from bec_lib.tests.utils import ConnectorMock
from bec_server.scan_bundler import ScanBundler

# overwrite threads_check fixture from bec_lib,
# to have it in autouse


@pytest.fixture(autouse=True)
def threads_check(threads_check):
    yield
    bec_logger.logger.remove()


class ScanBundlerMock(ScanBundler):
    def __init__(self, device_manager, connector_cls) -> None:
        super().__init__(
            ServiceConfig(redis={"host": "dummy", "port": 6379}), connector_cls=ConnectorMock
        )
        self.device_manager = device_manager

    def _start_device_manager(self):
        pass

    def _start_metrics_emitter(self):
        pass

    def _start_update_service_info(self):
        pass

    def wait_for_service(self, name, status=BECStatus.RUNNING):
        pass


dir_path = os.path.dirname(bec_lib.__file__)


@pytest.fixture
def scan_bundler_mock(dm_with_devices):
    device_manager = dm_with_devices
    scan_bundler_mock = ScanBundlerMock(device_manager, device_manager.connector)
    yield scan_bundler_mock
    scan_bundler_mock.shutdown()



from unittest import mock

import pytest

from bec_lib import messages
from bec_server.scan_bundler.emitter import EmitterBase


@pytest.mark.parametrize(
    "msgs",
    [
        ([]),
        (
            [
                (
                    messages.ScanMessage(point_id=1, scan_id="scan_id", data={}, metadata={}),
                    "endpoint",
                    None,
                )
            ]
        ),
        (
            [
                (
                    messages.ScanMessage(point_id=1, scan_id="scan_id", data={}, metadata={}),
                    "endpoint",
                    None,
                ),
                (
                    messages.ScanMessage(point_id=2, scan_id="scan_id", data={}, metadata={}),
                    "endpoint",
                    None,
                ),
            ]
        ),
        (
            [
                (
                    messages.ScanMessage(point_id=1, scan_id="scan_id", data={}, metadata={}),
                    "endpoint",
                    "public_endpoint",
                ),
                (
                    messages.ScanMessage(point_id=2, scan_id="scan_id", data={}, metadata={}),
                    "endpoint",
                    "public_endpoint",
                ),
            ]
        ),
    ],
)
def test_publish_data(msgs):
    connector = mock.MagicMock()
    with mock.patch.object(EmitterBase, "_start_buffered_connector") as start:
        emitter = EmitterBase(connector)
        start.assert_called_once()
        with mock.patch.object(emitter, "_get_messages_from_buffer", return_value=msgs) as get_msgs:
            emitter._publish_data()
            get_msgs.assert_called_once()

            if not msgs:
                connector.send.assert_not_called()
                return

            pipe = connector.pipeline()
            msgs_bundle = messages.BundleMessage()
            _, endpoint, _ = msgs[0]
            for msg, endpoint, public in msgs:
                msg_dump = msg
                msgs_bundle.append(msg_dump)
                if public:
                    connector.set.assert_has_calls(
                        connector.set(public, msg_dump, pipe=pipe, expire=1800)
                    )

            connector.send.assert_called_with(endpoint, msgs_bundle, pipe=pipe)
        emitter.shutdown()


@pytest.mark.parametrize(
    "msg,endpoint,public",
    [
        (
            messages.ScanMessage(point_id=1, scan_id="scan_id", data={}, metadata={}),
            "endpoint",
            None,
        ),
        (
            messages.ScanMessage(point_id=1, scan_id="scan_id", data={}, metadata={}),
            "endpoint",
            "public",
        ),
    ],
)
def test_add_message(msg, endpoint, public):
    connector = mock.MagicMock()
    emitter = EmitterBase(connector)
    emitter.add_message(msg, endpoint, public)
    msgs = emitter._get_messages_from_buffer()
    out_msg, out_endpoint, out_public = msgs[0]
    assert out_msg == msg
    assert out_endpoint == endpoint
    assert out_public == public
    emitter.shutdown()



from unittest import mock

from test_scibec_connector import SciHubMock

from bec_lib import MessageEndpoints, messages
from bec_server.scihub.scilog import SciLogConnector


def test_scilog_connector(SciHubMock):
    scilog = SciLogConnector(SciHubMock, SciHubMock.connector)
    scilog.shutdown()


def test_scilog_connector_with_env(SciHubMock):
    with mock.patch("bec_server.scihub.scilog.scilog.dotenv_values") as mock_dotenv_values:
        mock_dotenv_values.return_value = {
            "SCILOG_DEFAULT_HOST": "https://dummy_host",
            "SCILOG_USER": "dummy_user",
            "SCILOG_USER_SECRET": "dummy_user_secret",
        }
        SciHubMock.config.service_config = {"scilog": {"env_file": "dummy_env_file"}}
        with mock.patch("bec_server.scihub.scilog.scilog.requests") as mock_requests:
            mock_requests.post.return_value.ok = True
            mock_requests.post.return_value.json.return_value = {"token": "dummy_token"}
            scilog = SciLogConnector(SciHubMock, mock.MagicMock())
            assert scilog.host == "https://dummy_host"
            assert scilog.user == "dummy_user"
            assert scilog.user_secret == "dummy_user_secret"
            scilog.connector.set.assert_called_once_with(
                MessageEndpoints.logbook(),
                messages.CredentialsMessage(
                    credentials={
                        "url": "https://dummy_host",
                        "token": "Bearer dummy_token",
                        "user": "dummy_user",
                    }
                ),
            )
            scilog.shutdown()
            mock_dotenv_values.assert_called_once()



import builtins
from unittest import mock

import pytest

from bec_lib.callback_handler import EventType
from bec_lib.user_scripts_mixin import UserScriptsMixin

# pylint: disable=no-member
# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name
# pylint: disable=protected-access


def dummy_func():
    pass


def dummy_func2():
    pass


class client_user_scripts_mixin(UserScriptsMixin):
    def __init__(self):
        self.callbacks = None
        super().__init__()
        self._scripts = {}


@pytest.fixture
def scripts():
    yield client_user_scripts_mixin()


def test_user_scripts_forget(scripts):
    scripts.callbacks = mock.MagicMock()
    mock_run = scripts.callbacks.run
    scripts._scripts = {"test": {"cls": dummy_func, "file": "path_to_my_file.py"}}
    builtins.test = dummy_func
    scripts.forget_all_user_scripts()
    assert mock_run.call_count == 1
    assert mock_run.call_args == mock.call(
        EventType.NAMESPACE_UPDATE, action="remove", ns_objects={"test": dummy_func}
    )
    assert "test" not in builtins.__dict__
    assert len(scripts._scripts) == 0


def test_user_script_forget(scripts):
    scripts.callbacks = mock.MagicMock()
    mock_run = scripts.callbacks.run
    scripts._scripts = {"test": {"cls": dummy_func, "file": "path_to_my_file.py"}}
    builtins.test = dummy_func
    scripts.forget_user_script("test")
    assert mock_run.call_count == 1
    assert mock_run.call_args == mock.call(
        EventType.NAMESPACE_UPDATE, action="remove", ns_objects={"test": dummy_func}
    )
    assert "test" not in builtins.__dict__


def test_load_user_script(scripts):
    scripts.callbacks = mock.MagicMock()
    mock_run = scripts.callbacks.run
    builtins.__dict__["dev"] = scripts
    dummy_func.__module__ = "scripts"
    with mock.patch.object(scripts, "_run_linter_on_file") as linter:
        with mock.patch.object(
            scripts,
            "_load_script_module",
            return_value=[("test", dummy_func), ("wrong_test", dummy_func2)],
        ) as load_script:
            scripts.load_user_script("dummy")
            assert load_script.call_count == 1
            assert load_script.call_args == mock.call("dummy")
            assert "test" in scripts._scripts
            assert mock_run.call_count == 1
            assert mock_run.call_args == mock.call(
                EventType.NAMESPACE_UPDATE, action="add", ns_objects={"test": dummy_func}
            )
            assert "wrong_test" not in scripts._scripts
        # linter.assert_called_once_with("dummy") #TODO: re-enable this test once issue #298 is fixed


# def test_user_script_linter():
#     scripts = UserScriptsMixin()
#     current_path = pathlib.Path(__file__).parent.resolve()
#     script_path = os.path.join(current_path, "test_data", "user_script_with_bug.py")
#     builtins.__dict__["dev"] = scripts
#     with mock.patch("bec_lib.user_scripts_mixin.logger") as logger:
#         scripts._run_linter_on_file(script_path)
#         logger.error.assert_called_once()
